# Copyright (c) 2021 Watsen Networks.  All Rights Reserved.

_x='config-true-seq-nodes'
_w='config-false-prefixes'
_v='table-name-map'
_u='table-keys'
_t='app_ns'
_s='sztpd_meta'
_r='SELECT schema_name FROM information_schema.schemata;'
_q='cert'
_p='postgresql'
_o='Cannot delete: '
_n='[^/]*=[^/]*'
_m='[^:]*:'
_l='yang-library'
_k='" does not exist.'
_j=':tenants/tenant'
_i='opaque'
_h='global-root'
_g='db_ver'
_f='sqlite'
_e=':memory:'
_d='key'
_c='first'
_b='Parent node ('
_a='.singletons'
_Z='mysql'
_Y='" already exists.'
_X='Node "'
_W='sort-by'
_V='.*/'
_U='last'
_T=') does not exist.'
_S='_idx'
_R='_key'
_Q='after'
_P='before'
_O='jsob'
_N='insert'
_M='row_id'
_L='ssl'
_K=':'
_J='point'
_I='pid'
_H='obu_idx'
_G='singletons'
_F='=[^/]*'
_E=True
_D='='
_C=False
_B='/'
_A=None
import os,re,sys,json,base64,pickle,yangson,binascii,pkg_resources,sqlalchemy as sa
from enum import IntFlag
from urllib.parse import quote
from sqlalchemy.sql import and_
from urllib.parse import unquote
from dataclasses import dataclass
from sqlalchemy.sql import bindparam
from sqlalchemy.schema import CreateTable
from .  import db_utils
known_text=b'a secret message'
jsob_type=sa.JSON
class ContentType(IntFlag):CONFIG_TRUE=1;CONFIG_FALSE=2;CONFIG_ANY=3
@dataclass
class DatabasePath:data_path:str=_A;schema_path:str=_A;jsob_data_path:str=_A;jsob_schema_path:str=_A;table_name:str=_A;row_id:int=_A;inside_path:str=_A;path_segments:list=_A;jsob:dict=_A;node_ptr:dict=_A;prev_ptr:dict=_A
class NodeAlreadyExists(Exception):0
class NodeNotFound(Exception):0
class ParentNodeNotFound(Exception):0
class TooManyNodesFound(Exception):0
class InvalidResourceTarget(Exception):0
class CreateCallbackFailed(Exception):0
class ChangeCallbackFailed(Exception):0
class CreateOrChangeCallbackFailed(Exception):0
class DataAccessLayer:
	def __init__(A,db_url,cacert_param=_A,cert_param=_A,key_param=_A,yl_obj=_A,app_ns=_A,opaque=_A):
		F=yl_obj;E=key_param;D=cert_param;C=cacert_param;B=db_url;A.app_ns=_A;A.engine=_A;A.metadata=_A;A.leafrefs=_A;A.referers=_A;A.ref_stat_collectors=_A;A.global_root_id=_A;A.table_keys=_A;A.schema_path_to_real_table_name=_A;A.config_false_prefixes=_A;A.post_dal_callbacks=_A;A.config_true_obu_seq_nodes=_A
		if F is _A:A._init(B,C,D,E)
		else:A.app_ns=app_ns;A._create(B,C,D,E,F,opaque)
		assert A.app_ns!=_A;assert A.engine!=_A;assert A.metadata!=_A;assert A.leafrefs!=_A;assert A.referers!=_A;assert A.ref_stat_collectors!=_A;assert A.global_root_id!=_A;assert A.table_keys!=_A;assert A.schema_path_to_real_table_name!=_A;assert A.config_false_prefixes!=_A;assert A.config_true_obu_seq_nodes!=_A;A.len_app_ns=len(A.app_ns)
	def opaque(A):
		B=A.metadata.tables[A.schema_path_to_real_table_name[_G]]
		with A.engine.connect()as C:D=C.execute(sa.select([B.c.jsob]).where(B.c.name==_i));return D.first()[0]
	async def num_elements_in_list(A,data_path):
		B=data_path;C=re.sub(_F,'',B)
		if C!=B:assert NotImplementedError("Nested listed arren't supported yet.")
		D=A.schema_path_to_real_table_name[C];E=A.metadata.tables[D]
		with A.engine.connect()as F:G=F.execute(sa.select([sa.func.count()]).select_from(E));H=G.first()[0];return H
	async def get_tenant_name_for_admin(A,email_address):
		K='email-address';G=email_address
		if A.app_ns.endswith('x'):
			B=A.schema_path_to_real_table_name[_B+A.app_ns+':tenants/tenant/admin-accounts/admin-account'];C=A.metadata.tables[B]
			with A.engine.connect()as E:
				D=E.execute(sa.select([C.c.pid]).where(C.c[K]==G));F=D.first()
				if F!=_A:I=F[0];B=A.schema_path_to_real_table_name[_B+A.app_ns+_j];H=A.metadata.tables[B];D=E.execute(sa.select([H.c.name]).where(H.c.row_id==I));J=D.first()[0];return J
		B=A.schema_path_to_real_table_name[_B+A.app_ns+':admin-accounts/admin-account'];C=A.metadata.tables[B]
		with A.engine.connect()as E:
			D=E.execute(sa.select([C.c.pid]).where(C.c[K]==G));F=D.first()
			if F!=_A:return _A
		raise NodeNotFound('Admin "'+G+_k)
	async def get_tenant_name_for_global_key(A,table_name,k):
		B=table_name;C=A.schema_path_to_real_table_name[B];E=A.metadata.tables[C]
		with A.engine.connect()as F:
			D=F.execute(sa.select([E.c.pid]).where(getattr(E.c,A.table_keys[B])==k));G=D.first()
			if G==_A:raise NodeNotFound('key "'+k+'" in table "'+B+_k)
			I=G[0];C=A.schema_path_to_real_table_name[_B+A.app_ns+_j];H=A.metadata.tables[C];D=F.execute(sa.select([H.c.name]).where(H.c.row_id==I));J=D.first()[0];return J
	async def handle_get_opstate_request(B,data_path,query_dict={}):
		C=query_dict;A=data_path;assert A!='';assert not(A!=_B and A[-1]==_B)
		if A=='/ietf-yang-library:yang-library':
			F=B.schema_path_to_real_table_name[_G];D=B.metadata.tables[F]
			with B.engine.connect()as G:H=G.execute(sa.select([D.c.jsob]).where(D.c.name==_l));return H.first()[0]
		E=re.sub(_F,'',A)
		if 0:return await B._handle_get_data_request(A,E,ContentType.CONFIG_FALSE,C)
		else:return await B._handle_get_data_request(A,E,ContentType.CONFIG_ANY,C)
	async def handle_get_config_request(B,data_path,query_dict={}):
		E=data_path;A=re.sub(_F,'',E);F=await B._handle_get_data_request(E,A,ContentType.CONFIG_TRUE,query_dict);C=re.sub(_V,'',A)
		if not C.startswith(B.app_ns+_K):C=B.app_ns+_K+C
		for D in B.config_false_prefixes:
			if D.startswith(A):
				if A==_B:G=D[1:]
				else:G=D.replace(A,C,1)
				I=_A;J=F
				def H(prev_ptr,curr_ptr,remainder_path):
					D=prev_ptr;B=remainder_path;A=curr_ptr;E=B.split(_B)
					for C in E:
						if type(A)==list:
							for F in A:H(A,F,B)
							return
						elif C in A:D=A;A=A[C];B=B.replace(C+_B,'',1)
						else:return
					D.pop(C)
				H(I,J,G)
			else:0
		return F
	async def _handle_get_data_request(A,data_path,schema_path,content_type,query_dict):
		R='Node (';I=content_type;G=schema_path;F=data_path
		if I is ContentType.CONFIG_TRUE and any((G.startswith(B)for B in A.config_false_prefixes)):raise NodeNotFound(R+F+_T)
		with A.engine.connect()as J:
			B=A._get_dbpath_for_data_path(F,I,J)
			if B==_A:raise NodeNotFound(R+F+_T)
			if F!=_B and B.schema_path in A.schema_path_to_real_table_name:
				K,N=F.rsplit(_B,1)
				if _D not in N:
					if K=='':L=_B
					else:L=re.sub(_F,'',K)
					E={};await A._recursively_attempt_to_get_data_from_subtable(B.schema_path,B.row_id,L,B.jsob_data_path,E,I,query_dict,J)
					if E!={}:
						C=next(iter(E))
						if not C.startswith(A.app_ns+_K):E[A.app_ns+_K+C]=E[C];E.pop(C)
					else:
						C=G.rsplit(_B,1)[1]
						if not C.startswith(A.app_ns+_K):C=A.app_ns+_K+C
						E[C]=[]
					return E
			D=re.sub(_V,'',G);O=D
			if D=='':H=B.node_ptr
			else:
				if not D.startswith(A.app_ns+_K):D=A.app_ns+_K+D
				H={}
				if B.table_name!=_G and O==B.inside_path:H[D]=[];H[D].append(B.node_ptr);B.node_ptr=H[D][0]
				else:H[D]=B.node_ptr
			P=A._get_list_of_direct_subtables_for_schema_path(G)
			for M in P:Q=F+M[len(G):];await A._recursively_attempt_to_get_data_from_subtable(M,B.row_id,G,Q,B.node_ptr,I,{},J)
		return H
	async def _recursively_attempt_to_get_data_from_subtable(D,subtable_name,pid,jsob_schema_path,jsob_data_path,jsob_iter,content_type,query_dict,conn):
		K=query_dict;J=content_type;I=jsob_schema_path;C=subtable_name
		if not ContentType.CONFIG_FALSE in J:
			if any((C.startswith(A)for A in D.config_false_prefixes)):return
			else:0
		else:0
		if C in D.config_true_obu_seq_nodes:assert _W not in K;K[_W]=_H
		R=D._find_rows_in_table_having_pid(C,pid,K,conn);L=R.fetchall()
		if len(L)==0:return
		B=jsob_iter
		if I==_B:O=C[1:]
		else:assert I.startswith(_B+D.app_ns);O=C.replace(I+_B,'')
		E=re.sub(_V,'',C);G=re.sub(E+'$','',O)
		if G!='':
			G=G[:-1];S=G.split(_B)
			for M in S:
				try:B=B[M]
				except KeyError as X:assert ContentType.CONFIG_FALSE in J;B[M]={};B=B[M]
		if any((C.startswith(A)for A in D.config_false_prefixes)):
			B[E]=[];T=D.schema_path_to_real_table_name[C];U=D.metadata.tables[T]
			for F in L:
				H={}
				for A in U.c:
					if A.name!=_M and A.name!=_I:
						if type(A.type)is sa.sql.sqltypes.DateTime:H[A.name]=F[A.name].strftime('%Y-%m-%dT%H:%M:%SZ')
						elif type(A.type)is sa.sql.sqltypes.JSON or type(A.type)is sa.sql.sqltypes.PickleType:
							if F[A.name]is not _A and not(type(F[A.name])is list and len(F[A.name])==0):H[A.name]=F[A.name]
						elif F[A.name]is not _A:H[A.name]=F[A.name]
				B[E].append(H)
		else:
			P=0
			for N in L:
				Q=N[_O];E=next(iter(Q))
				if E not in B:B[E]=[]
				B[E].append(Q.pop(E));V=D._get_list_of_direct_subtables_for_schema_path(C)
				for W in V:await D._recursively_attempt_to_get_data_from_subtable(W,N[0],C,jsob_data_path+_D+N[2],B[E][P],J,{},conn)
				P+=1
	async def handle_post_config_request(A,data_path,query_dict,request_body,create_callbacks,change_callbacks,opaque):
		E=request_body;B=data_path;assert B!='';assert not(B!=_B and B[-1]==_B);assert'?'not in B;assert type(E)==dict
		with A.engine.begin()as C:
			F=A._get_dbpath_for_data_path(B,ContentType.CONFIG_TRUE,C)
			if F==_A:raise ParentNodeNotFound(_b+B+_T)
			await A._handle_post_config_request(F,query_dict,E,create_callbacks,change_callbacks,opaque,C)
			if A.post_dal_callbacks is not _A:
				for D in A.post_dal_callbacks:
					try:await D[0](D[1],C,D[2])
					except Exception as G:A.post_dal_callbacks=_A;raise G
				A.post_dal_callbacks=_A
	async def _handle_post_config_request(B,parent_dbpath,query_dict,request_body,create_callbacks,change_callbacks,opaque,conn):
		a=change_callbacks;V=opaque;U=create_callbacks;K=conn;J=query_dict;F=request_body;A=parent_dbpath;O=A.data_path;b=re.sub(_F,'',O);E=next(iter(F))
		if O==_B:C=E;L=_B+C
		else:C=re.sub(_m,'',E);L=O+_B+C
		M=re.sub(_F,'',L);H=B._get_table_name_for_schema_path(M);assert H!=_A
		if H==A.table_name:
			if C in A.node_ptr:
				if type(A.node_ptr[C])!=list:raise NodeAlreadyExists(_X+C+_Y)
				assert len(F[E])==1
				if F[E][0]in A.node_ptr[C]:raise NodeAlreadyExists(_X+F[E][0]+_Y)
				if M not in B.config_true_obu_seq_nodes:A.node_ptr[C].append(F[E][0])
				else:
					G=_U
					if _N in J:G=J[_N]
					if G==_c:assert _J not in J;A.node_ptr[C].insert(0,F[E][0])
					elif G==_U:assert _J not in J;A.node_ptr[C].append(F[E][0])
					else:
						assert G in(_P,_Q);assert _J in J;P=J[_J].rsplit(_D,1)[1];P=unquote(P);D=A.node_ptr[C].index(P)
						if G==_P:0
						elif G==_Q:D=D+1
						A.node_ptr[C].insert(D,F[E][0])
				if M in U:
					for Y in U[M]:await Y(L+_D+F[E][0],A.jsob,A.jsob_data_path,V)
			else:A.node_ptr[C]=F.pop(E);c=await B._recursively_post_subtable_data(A.row_id,L,A.node_ptr[C],A.jsob,A.jsob_data_path,U,V,K)
			B._update_jsob_for_row_id_in_table(A.table_name,A.row_id,A.jsob,K)
		else:
			assert M in B.table_keys
			if C not in A.node_ptr:A.node_ptr[C]=[];B._update_jsob_for_row_id_in_table(A.table_name,A.row_id,A.jsob,K)
			Q=F[E][0][B.table_keys[H]]
			if M in B.config_true_obu_seq_nodes:I=B._get_obu_idx_list_for_list_path(L,K)
			W={};assert len(F[E])==1;W[C]=F[E][0];R={};R[_I]=A.row_id;R[B.table_keys[H]]=Q
			if M in B.config_true_obu_seq_nodes:R[_H]=-1
			R[_O]={};S=B.schema_path_to_real_table_name[H];N=B.metadata.tables[S]
			try:Z=K.execute(N.insert().values(**R))
			except sa.exc.IntegrityError:raise NodeAlreadyExists(_X+C+_Y)
			L+=_D+str(Q);c=await B._recursively_post_subtable_data(Z.inserted_primary_key[0],L,W[C],W,L,U,V,K);S=B.schema_path_to_real_table_name[H];N=B.metadata.tables[S];K.execute(N.update().where(N.c.row_id==Z.inserted_primary_key[0]).values(jsob=W))
			if M in B.config_true_obu_seq_nodes:
				if O==_B:d=_B+C
				else:d=O+_B+C
				G=_U
				if _N in J:G=J[_N]
				if G==_c:D=0;T=[(B.table_keys[H],Q),(_H,D)];I.insert(0,T)
				elif G==_U:D=len(I);T=[(B.table_keys[H],Q),(_H,D)];I.append(T)
				else:
					assert G in(_P,_Q);P=J[_J].rsplit(_D,1)[1];D=0
					for X in I:
						if X[0][1]==P:break
						D+=1
					if G==_P:0
					elif G==_Q:D=D+1
					T=[(B.table_keys[H],Q),(_H,D)];I.insert(D,T)
				for X in range(D+1,len(I)):I[X][1]=_H,I[X][1][1]+1
				S=B.schema_path_to_real_table_name[H];N=B.metadata.tables[S];e=B.table_keys[H];f=N.update().where(and_(N.c.pid==A.row_id,getattr(N.c,e)==bindparam(_R))).values(obu_idx=bindparam(_S));Z=K.execute(f,[{_R:A[0][1],_S:A[1][1]}for A in I[D:len(I)]])
			else:0
		g=re.sub(_F,'',O)
		if g in a:
			for Y in a[b]:await Y(O,A.jsob,A.jsob_data_path,V)
	async def _recursively_post_subtable_data(B,pid,data_path,req_body_iter,jsob,jsob_data_path,create_callbacks,opaque,conn):
		P=pid;M=jsob_data_path;L=jsob;I=conn;H=opaque;E=create_callbacks;D=data_path;A=req_body_iter;C=re.sub(_F,'',D)
		if type(A)==dict:
			if C in E:
				for Q in E[C]:await Q(D,L,M,H)
			for N in A.copy():
				if N.startswith(B.app_ns):F=D+_B+N[B.len_app_ns+1:]
				else:F=D+_B+N
				R=await B._recursively_post_subtable_data(P,F,A[N],L,M,E,H,I)
		elif type(A)==list:
			if C in B.table_keys:
				if C in B.config_true_obu_seq_nodes:S=0
				while A:
					J=A.pop(0);assert type(J)==dict;V=re.sub(_V,'',C);O={};O[V]=J;K=B.table_keys[C];G={};G[_I]=P;G[K]=J[K]
					if C in B.config_true_obu_seq_nodes:G[_H]=S;S+=1
					G[_O]=O;W=B.schema_path_to_real_table_name[C];X=B.metadata.tables[W]
					try:T=I.execute(X.insert().values(**G))
					except sa.exc.IntegrityError as Y:raise NodeAlreadyExists(_X+K+'" with value "'+G[K]+_Y)
					F=D+_D+str(J[K]);R=await B._recursively_post_subtable_data(T.inserted_primary_key[0],F,J,O,F,E,H,I);B._update_jsob_for_row_id_in_table(C,T.inserted_primary_key[0],O,I)
				assert type(A)==list;assert len(A)==0
			else:
				assert type(A)==list
				if not(len(A)==1 and A[0]==_A):
					for U in A:F=D+_D+str(U);R=await B._recursively_post_subtable_data(P,F,A[A.index(U)],L,M,E,H,I)
				else:0
		elif C in E:
			for Q in E[C]:await Q(D,L,M,H)
		if C in B.table_keys:return _E
		return _C
	async def handle_post_opstate_request(A,data_path,request_body):
		P='Unrecognized resource schema path: ';I=request_body;E=data_path;B=re.sub(_F,'',E);F=next(iter(I))
		if B==_B:G=F;C=_B+G
		else:G=re.sub(_m,'',F);C=B+_B+G
		J=A._get_table_name_for_schema_path(C)
		if J!=C:raise NodeNotFound(P+C)
		D=I[F];K=re.findall(_n,E)
		if len(K)==0:D[_I]=A.global_root_id
		else:
			L=A._get_table_name_for_schema_path(B)
			if L==_A:raise ParentNodeNotFound(P+B)
			M=K[-1].split(_D)
			with A.engine.connect()as H:D[_I]=A._get_row_id_for_key_in_table(L,M[1],H)
			if D[_I]==_A:raise ParentNodeNotFound('Nonexistent parent resource: '+E)
		N=A.schema_path_to_real_table_name[J];O=A.metadata.tables[N]
		with A.engine.connect()as H:Q=H.execute(O.insert().values(**D))
	async def handle_put_config_request(A,data_path,query_dict,request_body,create_callbacks={},change_callbacks={},delete_callbacks={},opaque=_A):
		B=data_path;assert B!='';assert not(B!=_B and B[-1]==_B)
		with A.engine.begin()as D:
			E=await A._handle_put_config_request(B,query_dict,request_body,create_callbacks,change_callbacks,delete_callbacks,opaque,D)
			if A.post_dal_callbacks is not _A:
				for C in A.post_dal_callbacks:
					try:await C[0](C[1],D,C[2])
					except Exception as F:A.post_dal_callbacks=_A;raise F
				A.post_dal_callbacks=_A
			return E
	async def _handle_put_config_request(B,data_path,query_dict,request_body,create_callbacks,change_callbacks,delete_callbacks,opaque,conn):
		U=delete_callbacks;S=create_callbacks;P=opaque;L=change_callbacks;I=data_path;F=query_dict;D=request_body;C=conn;K=re.sub(_F,'',I);assert type(D)==dict
		if I==_B:A=B._get_dbpath_for_data_path(_B,ContentType.CONFIG_ANY,C);assert A!=_A;await B.recursive_compare_and_put(A.row_id,_B,D,A.node_ptr,_A,A,S,L,U,P,C);M=B.schema_path_to_real_table_name[A.table_name];J=B.metadata.tables[M];C.execute(J.update().where(J.c.row_id==A.row_id).values(jsob=A.jsob));return _C
		assert len(D)==1;assert K!=_B;A=B._get_dbpath_for_data_path(I,ContentType.CONFIG_ANY,C)
		if A==_A:
			assert I!=_B;G,X=I.rsplit(_B,1)
			if G=='':G=_B
			A=B._get_dbpath_for_data_path(G,ContentType.CONFIG_ANY,C)
			if A==_A:raise ParentNodeNotFound(_b+G+_T)
			await B._handle_post_config_request(A,F,D,S,L,P,C);B._update_jsob_for_row_id_in_table(A.table_name,A.row_id,A.jsob,C);return _E
		Y=next(iter(D));D=D[Y]
		if type(D)==list:assert len(D)==1;D=D[0]
		await B.recursive_compare_and_put(A.row_id,I,D,A.node_ptr,A.prev_ptr,A,S,L,U,P,C)
		if K in B.config_true_obu_seq_nodes and F!=_A and _N in F:
			if type(A.prev_ptr)==list:A.prev_ptr.remove(A.node_ptr)
			else:assert type(A.prev_ptr)==dict;Z,Q=I.rsplit(_D,1);H=B._get_obu_idx_list_for_list_path(Z,C);H=[A for A in H if A[0][1]!=Q]
			N=F[_N]
			if N==_c:
				assert _J not in F
				if type(A.prev_ptr)==list:A.prev_ptr.insert(0,A.node_ptr)
				else:E=0;O=[(B.table_keys[K],Q),(_H,E)];H.insert(0,O)
			elif N==_U:
				assert _J not in F
				if type(A.prev_ptr)==list:A.prev_ptr.append(A.node_ptr)
				else:E=len(H);O=[(B.table_keys[K],Q),(_H,E)];H.append(O)
			else:
				assert N in(_P,_Q);assert _J in F;V=unquote(F[_J].rsplit(_D,1)[1])
				if type(A.prev_ptr)==list:E=A.prev_ptr.index(V)
				else:
					E=0
					for R in H:
						if R[0][1]==V:break
						E+=1
				if N==_P:0
				elif N==_Q:E=E+1
				if type(A.prev_ptr)==list:A.prev_ptr.insert(E,A.node_ptr)
				else:O=[(B.table_keys[K],Q),(_H,E)];H.insert(E,O)
			if type(A.prev_ptr)!=list:
				for R in range(len(H)):H[R][1]=_H,R
				M=B.schema_path_to_real_table_name[K];J=B.metadata.tables[M];a=B.table_keys[K];b=B._get_row_data_for_list_path(I,C);c=J.update().where(and_(J.c.pid==b[_I],getattr(J.c,a)==bindparam(_R))).values(obu_idx=bindparam(_S));e=C.execute(c,[{_R:A[0][1],_S:A[1][1]}for A in H])
		M=B.schema_path_to_real_table_name[A.table_name];J=B.metadata.tables[M];C.execute(J.update().where(J.c.row_id==A.row_id).values(jsob=A.jsob))
		if K in B.config_true_obu_seq_nodes and F!=_A and _N in F:
			G,X=I.rsplit(_B,1)
			if G=='':G=_B
			W=re.sub(_F,'',G)
			if W in L:
				T=B._get_dbpath_for_data_path(G,ContentType.CONFIG_TRUE,C);assert T!=_A
				for d in L[W]:await d(G,T.jsob,T.jsob_data_path,P)
		return _C
	async def recursive_compare_and_put(C,pid,data_path,req_body_iter,dbpath_curr_ptr,dbpath_prev_ptr,dbpath,create_callbacks,change_callbacks,delete_callbacks,opaque,conn):
		Y=dbpath_prev_ptr;P=pid;N=delete_callbacks;M=change_callbacks;L=create_callbacks;J=opaque;I=conn;H=dbpath;F=dbpath_curr_ptr;D=req_body_iter;B=data_path;assert type(D)==type(F);E=re.sub(_F,'',B)
		if B==_B:0
		if type(D)==dict:
			S=set(list(D.keys()));T=set(list(F.keys()))
			for A in [A for A in S if A not in T]:
				if type(D[A])==list:
					U=_B+A if B==_B else B+_B+A;V=re.sub(_F,'',U)
					if V in C.table_keys:await C.recursive_compare_and_put(P,U,D[A],[],_A,H,L,M,N,J,I);F[A]=[]
					else:F[A]=[];await C.recursive_compare_and_put(P,U,D[A],F[A],F,H,L,M,N,J,I)
				else:
					E=re.sub(_F,'',B);V=_B+A if E==_B else E+_B+A;assert type(D)!=list;G=_B+A if B==_B else B+_B+A;F[A]=D[A];d=await C._recursively_post_subtable_data(P,G,D[A],H.jsob,H.jsob_data_path,L,J,I)
					if d==1:assert type(D)==dict;assert type(D[A])==list;assert len(D[A])==0;D.pop(A);F.pop(A)
			for A in T-S:
				if _C:U=_B+A if B==_B else B+_B+A;await C.recursive_compare_and_put(P,U,[],F[A],F,H,L,M,N,J,I);del F[A]
				else:
					V=_B+A if E==_B else E+_B+A
					if V in C.config_false_prefixes:0
					else:G=_B+A if B==_B else B+_B+A;await C._recursively_delete_subtable_data(H.row_id,G,F[A],N,J,I);del F[A]
			Z=_C
			for A in T&S:
				G=_B+A if B==_B else B+_B+A;V=re.sub(_F,'',G);e=await C.recursive_compare_and_put(P,G,D[A],F[A],F,H,L,M,N,J,I)
				if e==_E:Z=_E
			if T-S or(S-T or Z==_E):
				if E in M:
					for R in M[E]:await R(B,H.jsob,H.jsob_data_path,J)
			return _C
		elif type(D)==list:
			if E in C.table_keys:
				W=[A[C.table_keys[E]]for A in D];O=set(W);K=set([A[0]for A in C._get_keys_in_table_having_pid(E,P,I)])
				for A in [A for A in W if A not in K]:assert B!=_B;G=B+_D+A;f=[B for B in D if B[C.table_keys[E]]==A][0];g=[f];n=await C._recursively_post_subtable_data(P,B,g,_A,_A,L,J,I)
				for A in K-O:G=B+_D+A;o,a=B.rsplit(_B,1);assert a!='';await C._recursively_delete_subtable_data(H.row_id,G,Y[a],N,J,I)
				for A in K&O:G=B+_D+A;Q=C._get_dbpath_for_data_path(G,ContentType.CONFIG_TRUE,I);assert Q!=_A;h=[B for B in D if B[C.table_keys[E]]==A][0];await C.recursive_compare_and_put(Q.row_id,G,h,Q.node_ptr,Q.prev_ptr,Q,L,M,N,J,I);C._update_jsob_for_row_id_in_table(Q.table_name,Q.row_id,Q.jsob,I)
				b=_C
				if E in C.config_true_obu_seq_nodes:
					i=C._get_obu_idx_list_for_list_path(B,I);c=[[(_d,A),(_H,W.index(A))]for A in W]
					if c!=i:j=C.schema_path_to_real_table_name[E];X=C.metadata.tables[j];k=C.table_keys[E];l=X.update().where(and_(X.c.pid==P,getattr(X.c,k)==bindparam(_R))).values(obu_idx=bindparam(_S));p=I.execute(l,[{_R:A[0][1],_S:A[1][1]}for A in c]);b=_E
				if O-K or K-O or b==_E:return _E
				return _C
			else:
				O=set(D);K=set(F);F.clear();F.extend(D)
				for A in O-K:
					G=B+_D+unquote(A)
					if E in L:
						for R in L[E]:await R(G,H.jsob,H.jsob_data_path,J)
				for A in K-O:
					G=B+_D+unquote(A)
					if E in N:
						for R in N[E]:await R(G,J)
				if O-K or K-O:return _E
				return _C
		else:
			if F!=D:
				m=re.sub('^.*/','',B);Y[m]=D
				if E in M:
					for R in M[E]:await R(B,H.jsob,H.jsob_data_path,J)
			else:0
			return _C
		raise NotImplementedError('logic never reaches this point')
	async def handle_put_opstate_request(D,data_path,request_body):
		F=request_body;B=data_path;assert B!='';assert not(B!=_B and B[-1]==_B)
		with D.engine.begin()as G:
			C,E=B.rsplit(_B,1);assert E!=''
			if C=='':C=_B
			A=D._get_dbpath_for_data_path(C,ContentType.CONFIG_ANY,G)
			if A==_A:raise ParentNodeNotFound(_b+C+_T)
			assert type(A.node_ptr[E])==type(F);A.node_ptr[E]=F;D._update_jsob_for_row_id_in_table(A.table_name,A.row_id,A.jsob,G)
	async def handle_delete_config_request(A,data_path,delete_callbacks,change_callbacks,opaque):
		B=data_path;assert B!='';assert B!=_B;assert B[-1]!=_B
		with A.engine.begin()as D:
			await A._handle_delete_config_request(B,delete_callbacks,change_callbacks,opaque,D)
			if A.post_dal_callbacks is not _A:
				for C in A.post_dal_callbacks:
					try:await C[0](C[1],D,C[2])
					except Exception as E:A.post_dal_callbacks=_A;raise E
				A.post_dal_callbacks=_A
	async def _handle_delete_config_request(C,data_path,delete_callbacks,change_callbacks,opaque,conn):
		I=opaque;H=change_callbacks;F=conn;D=data_path;E,G=D.rsplit(_B,1)
		if E=='':E=_B
		A=C._get_dbpath_for_data_path(E,ContentType.CONFIG_TRUE,F)
		if A==_A:raise NodeNotFound(_o+D)
		if _D in G:B,O=G.rsplit(_D,1)
		else:B=G
		assert type(A.node_ptr)==dict
		if B not in A.node_ptr:raise NodeNotFound('Cannot delete '+D+'.')
		await C._recursively_delete_subtable_data(A.row_id,D,A.node_ptr[B],delete_callbacks,I,F)
		if type(A.node_ptr[B])==list:
			J=re.sub(_F,'',D)
			if J in C.table_keys:
				L=C._find_rowids_in_table_having_pid(J,A.row_id,F);M=L.fetchall()
				if len(M)==0:assert type(A.node_ptr[B])==list;assert len(A.node_ptr[B])==0;A.node_ptr.pop(B)
			elif len(A.node_ptr[B])==0:A.node_ptr.pop(B)
		else:A.node_ptr.pop(B)
		C._update_jsob_for_row_id_in_table(A.table_name,A.row_id,A.jsob,F);K=re.sub(_F,'',E)
		if K in H:
			for N in H[K]:await N(E,A.jsob,A.jsob_data_path,I)
	async def _recursively_delete_subtable_data(A,pid,data_path,curr_data_iter,delete_callbacks,opaque,conn):
		J=opaque;G=conn;F=delete_callbacks;E=pid;C=data_path;B=curr_data_iter;H=re.sub(_F,'',C)
		if type(B)==list:
			if H in A.table_keys:
				assert B==[];O,K=C.rsplit(_B,1)
				async def L(pid,data_path,delete_callbacks,opaque,conn):
					D=conn;C=data_path;F=re.sub(_F,'',C);B=A._get_dbpath_for_data_path(C,ContentType.CONFIG_TRUE,D)
					if B==_A:raise NodeNotFound(_o+C)
					G=next(iter(B.jsob));H=B.jsob[G];await A._recursively_delete_subtable_data(B.row_id,C,H,delete_callbacks,opaque,D);I=A.schema_path_to_real_table_name[F];E=A.metadata.tables[I];J=D.execute(sa.delete(E).where(E.c.row_id==B.row_id));assert J.rowcount==1
				if _D in K:await L(E,C,F,J,G)
				else:
					P=[B[0]for B in A._get_keys_in_table_having_pid(H,E,G)]
					for D in P:I=C+_D+D;await L(E,I,F,J,G)
			elif any((H.startswith(B)for B in A.config_false_prefixes)):assert B==[];Q=A.schema_path_to_real_table_name[H];M=A.metadata.tables[Q];S=G.execute(sa.delete(M).where(M.c.pid==E))
			else:
				O,K=C.rsplit(_B,1)
				if _D in K:D=unquote(K.rsplit(_D)[1]);assert D in B;I=C;await A._recursively_delete_subtable_data(E,I,D,F,J,G);B.remove(D)
				else:
					while len(B)!=0:
						D=B[0]
						if D is _A:break
						I=C+_D+D;await A._recursively_delete_subtable_data(E,I,D,F,J,G);B.pop(0)
		elif type(B)==dict:
			for N in B.keys():assert C!=_B;I=C+_B+N;await A._recursively_delete_subtable_data(E,I,B[N],F,J,G)
		else:0
		if not type(B)==list and not any((H.startswith(B)for B in A.config_false_prefixes)):
			if H in F:
				for R in F[H]:await R(C,J)
	def _find_rows_in_table_having_pid(E,table_name,pid,query_dict,conn):
		J='limit';I='offset';H='direction';B=query_dict;F=E.schema_path_to_real_table_name[table_name];C=E.metadata.tables[F];A=sa.select([C]).where(C.c.pid==pid)
		if _W in B:D=getattr(C.c,B[_W])
		else:D=C.c.row_id
		if H in B and B[H]=='backwards':A=A.order_by(D.desc())
		else:A=A.order_by(D.asc())
		if I in B:A=A.offset(int(B[I]))
		if J in B:A=A.limit(int(B[J]))
		G=conn.execute(A);return G
	def _find_rowids_in_table_having_pid(B,table_name,pid,conn):C=B.schema_path_to_real_table_name[table_name];A=B.metadata.tables[C];D=conn.execute(sa.select([A.c.row_id]).where(A.c.pid==pid).order_by(A.c.row_id));return D
	def _get_keys_in_table_having_pid(A,table_name,pid,conn):B=table_name;D=A.schema_path_to_real_table_name[B];C=A.metadata.tables[D];E=conn.execute(sa.select([getattr(C.c,A.table_keys[B])]).where(C.c.pid==pid));return E
	def _get_list_of_direct_subtables_for_schema_path(D,schema_path):
		A=schema_path
		if A!=_B:assert A[-1]!=_B;A+=_B
		C=[]
		for B in sorted(D.schema_path_to_real_table_name.keys()):
			if str(B).startswith(A):
				if not any((A for A in C if str(B).startswith(A+_B))):
					if str(B)!=_B:C.append(str(B))
					else:0
				else:0
		return C
	def _get_row_id_for_key_in_table(B,table_name,key,conn):
		C=table_name;E=B.schema_path_to_real_table_name[C];D=B.metadata.tables[E];F=conn.execute(sa.select([D.c.row_id]).where(getattr(D.c,B.table_keys[C])==key));A=F.fetchall();assert A is not _A
		if len(A)==0:return _A
		if len(A)>1:raise TooManyNodesFound()
		return A[0][0]
	def _get_jsob_iter_for_path_in_jsob(D,jsob,path):
		B=path;assert jsob!=_A;assert B[0]!=_B;A=jsob
		if B!='':
			for C in B.split(_B):
				if C!=''and C not in A:return _A
				A=A[C]
				if type(A)==list:assert len(A)==1;A=A[0]
		return A
	def _get_jsob_for_row_id_in_table(C,table_name,row_id,conn):
		F=row_id;D=table_name;K=C.schema_path_to_real_table_name[D];A=C.metadata.tables[K]
		if D in C.table_keys:E=conn.execute(sa.select([A.c.jsob]).where(A.c.row_id==F));G=E.first();assert G!=_A;return G[0]
		else:
			E=conn.execute(sa.select([A]).where(A.c.row_id==F));H=E.first();assert H!=_A;I=D.rsplit(_B,1)[1];J={I:{}}
			for B in A.c:
				if B.name!=_M and B.name!=_I:J[I][B.name]=H[B.name]
			return J
	def _insert_jsob_into_table(B,pid,table_name,new_jsob,conn,obu_idx=_A):
		F=obu_idx;D=table_name;C=new_jsob;H=B.schema_path_to_real_table_name[D];I=B.metadata.tables[H];E=next(iter(C));A={};A[_I]=pid
		if D in B.table_keys:
			A[B.table_keys[D]]=C[E][B.table_keys[D]]
			if F!=_A:A[_H]=F
			A[_O]=C
		else:
			for G in C[E].keys():A[G]=C[E][G]
		J=conn.execute(I.insert().values(**A));return J.inserted_primary_key[0]
	def _update_jsob_for_row_id_in_table(A,table_name,row_id,new_jsob,conn):C=A.schema_path_to_real_table_name[table_name];B=A.metadata.tables[C];D=conn.execute(sa.update(B).where(B.c.row_id==row_id).values(jsob=new_jsob))
	def _get_table_name_for_schema_path(D,schema_path):
		B=len(_B);C=_G
		for A in D.schema_path_to_real_table_name.keys():
			if schema_path.startswith(A)and len(A)>B:B=len(A);C=A
		return C
	def _get_obu_idx_list_for_list_path(B,list_path,conn):
		D=list_path;assert D[0]==_B;assert D!=_B;assert D[-1]!=_B;C='';E=B.global_root_id;K=D[1:].split(_B)
		for F in K:
			if _D in F:
				L,M=F.split(_D);C+=_B+L;I=B._get_table_name_for_schema_path(C);G=B.schema_path_to_real_table_name[I];A=B.metadata.tables[G];J=conn.execute(sa.select([A.c.row_id,A.c.pid]).where(and_(A.c.pid==E,getattr(A.c,B.table_keys[I])==M)));H=J.fetchone()
				if H==_A:return _A
				assert J.fetchone()==_A;assert E==H[_I];E=H[_M]
			else:C+=_B+F
		G=B.schema_path_to_real_table_name[C];A=B.metadata.tables[G];N=conn.execute(sa.select([getattr(A.c,B.table_keys[C]),A.c.obu_idx]).where(A.c.pid==E).order_by(A.c.obu_idx));O=N.fetchall();P=[A.items()for A in O];return P
	def _get_row_data_for_list_path(A,data_path,conn):
		B=data_path;assert B[0]==_B;assert B!=_B;assert B[-1]!=_B;G=B[1:].split(_B);assert _D in G[-1];D='';H=A.global_root_id
		for E in G:
			if _D in E:
				K,L=E.split(_D);D+=_B+K;I=A._get_table_name_for_schema_path(D);M=A.schema_path_to_real_table_name[I];C=A.metadata.tables[M];J=conn.execute(sa.select([C.c.row_id,C.c.pid]).where(and_(C.c.pid==H,getattr(C.c,A.table_keys[I])==L)));F=J.fetchone()
				if F==_A:return _A
				assert J.fetchone()==_A;H=F[_M]
			else:D+=_B+E
		return F
	def _get_dbpath_for_data_path(B,data_path,content_type,conn):
		D=conn;C=data_path;A=DatabasePath();A.data_path=C;A.schema_path=re.sub(_F,'',C);A.table_name=B._get_table_name_for_schema_path(A.schema_path)
		if A.table_name==_A:return _A
		if C!=_B and A.table_name==A.schema_path:
			G,N=C.rsplit(_B,1)
			if _D not in N:
				if G=='':H=_B
				else:H=re.sub(_F,'',G)
				A.table_name=B._get_table_name_for_schema_path(H);assert A.table_name!=_A
		if A.table_name==_G:A.jsob_data_path=_B;A.jsob_schema_path=_B
		else:
			A.jsob_data_path=C;A.jsob_schema_path=re.sub(_F,'',A.jsob_data_path)
			while A.jsob_schema_path!=A.table_name and A.jsob_schema_path!=_B:O=A.jsob_data_path;A.jsob_data_path=re.sub('(.*=[^/]*)/.*','\\g<1>',A.jsob_data_path);assert A.jsob_data_path!=O;A.jsob_schema_path=re.sub(_F,'',A.jsob_data_path)
			assert A.jsob_schema_path!=_B
		if ContentType.CONFIG_FALSE in content_type and any((A.table_name.startswith(C)for C in B.config_false_prefixes)):raise InvalidResourceTarget("RFC 8040 does not allow queries on lists directly and, because SZTPD doesn't support keys on 'config false' lists, it is never possible to query for 'dbpath.table_name' to be returned.  The 'val' layer should've rejected this query... ")
		if A.jsob_schema_path==_B:A.row_id=B.global_root_id
		else:
			assert A.jsob_schema_path in B.table_keys;assert _D in A.jsob_data_path;I=A.jsob_data_path.split(_B);assert _D in I[-1];E=I[-1].split(_D)
			try:A.row_id=B._get_row_id_for_key_in_table(A.table_name,E[1],D)
			except TooManyNodesFound:
				J=B._get_row_data_for_list_path(A.jsob_data_path,D)
				if J is _A:A.row_id=_A
				else:A.row_id=J[_M]
			if A.row_id==_A:return _A
		assert A.data_path.startswith(A.jsob_data_path);K=A.data_path[len(A.jsob_data_path):];assert A.schema_path.startswith(A.jsob_schema_path);L=A.schema_path[len(A.jsob_schema_path):]
		if A.table_name==_G:A.inside_path=A.schema_path[1:]
		else:M=re.findall(_n,A.jsob_data_path);assert len(M)!=0;E=M[-1].split(_D);A.inside_path=E[0];P=re.sub('^'+A.table_name,'',A.schema_path);assert P==L;A.inside_path+=L
		assert A.inside_path==''or A.inside_path[0]!=_B;A.jsob=B._get_jsob_for_row_id_in_table(A.table_name,A.row_id,D);A.node_ptr=A.jsob;A.prev_ptr=_A
		if A.inside_path=='':A.path_segments=[''];return A
		A.path_segments=A.inside_path.split(_B);Q=''
		for F in A.path_segments:
			Q+=_B+F
			if type(A.node_ptr)==list:raise NotImplementedError('is this path used?');A.prev_ptr=A.node_ptr;A.node_ptr=A.node_ptr[0]
			if F not in A.node_ptr:return _A
			else:A.prev_ptr=A.node_ptr;A.node_ptr=A.node_ptr[F]
		if type(A.node_ptr)==list:
			if _D in K:
				R=unquote(K.rsplit(_D,1)[1])
				try:S=A.node_ptr.index(R)
				except:return _A
				A.prev_ptr=A.node_ptr;A.node_ptr=A.node_ptr[S]
		return A
	def _init(A,url,cacert_param,cert_param,key_param):
		J=key_param;I=cert_param;H=cacert_param;G=url
		if not(G.startswith('sqlite:///')or G.startswith(_Z)or G.startswith(_p)):raise SyntaxError('The database url contains an unrecognized dialect.')
		F={}
		if H is not _A:
			F[_L]={};F[_L]['ca']=H
			if I is not _A:F[_L][_q]=I
			if J is not _A:F[_L][_d]=J
		A.engine=sa.create_engine(G,connect_args=F);A.db_schema=_A;A.table_keys={};A.config_false_prefixes={};A.config_true_obu_seq_nodes={};A.schema_path_to_real_table_name={};A.leafrefs={};A.referers={};A.ref_stat_collectors={}
		if A.engine.url.database==_e or not db_utils.database_exists(A.engine.url,connect_args=F):A.engine=_A;raise NotImplementedError
		if A.engine.dialect.name==_f:A.schema_path_to_real_table_name[_B]=_G;A.schema_path_to_real_table_name[_G]=_G
		else:
			A.db_schema=A.engine.url.database;A.schema_path_to_real_table_name[_B]=A.db_schema.join(_a);A.schema_path_to_real_table_name[_G]=A.db_schema+_a;M=A.engine.execute(_r);K=M.fetchall();N=[K[A][0]for A in range(len(K))]
			if A.db_schema not in N:A.engine.execute(sa.schema.CreateSchema(A.db_schema));raise NotImplementedError
		A.metadata=sa.MetaData(bind=A.engine,schema=A.db_schema);A.metadata.reflect()
		for O in A.metadata.tables.values():
			for D in O.c:
				if type(D.type)is sa.sql.sqltypes.BLOB or type(D.type)is sa.sql.sqltypes.PickleType:D.type=sa.PickleType()
				if A.engine.dialect.name==_Z and type(D.type)is sa.dialects.mysql.types.LONGTEXT:D.type=sa.JSON()
				elif type(D.type)is sa.sql.sqltypes.JSON:D.type=sa.JSON()
		B=A.metadata.tables[A.schema_path_to_real_table_name[_G]]
		with A.engine.connect()as E:
			C=E.execute(sa.select([B.c.jsob]).where(B.c.name==_s));L=C.first()[0]
			if L[_g]!=1:raise AssertionError('The database version ('+L[_g]+') is unexpected.')
			C=E.execute(sa.select([B.c.jsob]).where(B.c.name==_t));A.app_ns=C.first()[0];C=E.execute(sa.select([B.c.row_id]).where(B.c.name==_h));A.global_root_id=C.first()[0];C=E.execute(sa.select([B.c.jsob]).where(B.c.name==_u));A.table_keys=C.first()[0];C=E.execute(sa.select([B.c.jsob]).where(B.c.name==_v));A.schema_path_to_real_table_name=C.first()[0];C=E.execute(sa.select([B.c.jsob]).where(B.c.name==_w));A.config_false_prefixes=C.first()[0];C=E.execute(sa.select([B.c.jsob]).where(B.c.name==_x));A.config_true_obu_seq_nodes=C.first()[0]
	def _create(A,url,cacert_param,cert_param,key_param,yl_obj,opaque):
		V='default startup endpoint';U='endpoint';T='SZTPD_TEST_PATH';S='_sztp_ref_stats_stmt';R='_sztp_globally_unique_stmt';J=key_param;I=cert_param;H=cacert_param;G='name';F=yl_obj;E={}
		if H is not _A:
			E[_L]={};E[_L]['ca']=H
			if I is not _A:E[_L][_q]=I
			if J is not _A:E[_L][_d]=J
		A.engine=sa.create_engine(url,connect_args=E)
		if A.engine.url.database!=_e and db_utils.database_exists(A.engine.url,connect_args=E):raise AssertionError('Database already exists (call init() first).')
		if A.engine.url.database!=_e:
			if A.engine.dialect.name==_Z:db_utils.create_database(A.engine.url,encoding='utf8mb4',connect_args=E)
			else:db_utils.create_database(A.engine.url,connect_args=E)
		A.db_schema=_A
		if A.engine.dialect.name in(_Z,_p):
			A.db_schema=str(A.engine.url.database);D=A.engine.execute(_r);K=D.fetchall();M=[K[A][0]for A in range(len(K))]
			if A.db_schema not in M:A.engine.execute('CREATE SCHEMA IF NOT EXISTS %s;'%A.db_schema)
		A.metadata=sa.MetaData(bind=A.engine,schema=A.db_schema);C=sa.Table(_G,A.metadata,sa.Column(_M,sa.Integer,primary_key=_E),sa.Column(_I,sa.Integer,unique=_E),sa.Column(G,sa.String(250),unique=_E),sa.Column(_O,jsob_type));A.metadata.create_all()
		with A.engine.connect()as B:B.execute(C.insert(),name=_s,jsob={_g:1});B.execute(C.insert(),name=_t,jsob=A.app_ns);B.execute(C.insert(),name=_l,jsob=F);B.execute(C.insert(),name=_i,jsob=opaque)
		A.table_keys={_G:G,_B:G};A.config_true_obu_seq_nodes={};A.config_false_prefixes={};A.schema_path_to_real_table_name={};A.leafrefs={};A.referers={};A.ref_stat_collectors={}
		if A.engine.dialect.name==_f:A.schema_path_to_real_table_name[_B]=_G;A.schema_path_to_real_table_name[_G]=_G
		else:A.schema_path_to_real_table_name[_B]=A.db_schema+_a;A.schema_path_to_real_table_name[_G]=A.db_schema+_a
		def N(self,stmt,sctx):self.globally_unique=stmt.argument
		setattr(yangson.schemanode.SchemaNode,R,N);yangson.schemanode.SchemaNode._stmt_callback['wn-app:globally-unique']=R
		def O(self,stmt,sctx):self.ref_stats=stmt.argument
		setattr(yangson.schemanode.SchemaNode,S,O);yangson.schemanode.SchemaNode._stmt_callback['wn-app:ref-stats']=S;L=pkg_resources.resource_filename('sztpd','yang/')
		if os.environ.get(T):A.dm=yangson.DataModel(json.dumps(F),[os.environ.get(T),L])
		else:A.dm=yangson.DataModel(json.dumps(F),[L])
		A._gen_tables(A.dm.schema,_G)
		with A.engine.connect()as B:D=B.execute(C.insert(),name=_v,jsob=A.schema_path_to_real_table_name);D=B.execute(C.insert(),name=_u,jsob=A.table_keys);D=B.execute(C.insert(),name=_w,jsob=A.config_false_prefixes);D=B.execute(C.insert(),name=_x,jsob=A.config_true_obu_seq_nodes)
		if os.environ.get('SZTPD_TEST_DAL'):
			with A.engine.connect()as B:D=B.execute(C.insert().values(name=_h,jsob={}))
			A.global_root_id=D.inserted_primary_key[0]
		else:
			with A.engine.connect()as B:D=B.execute(C.insert().values(name=_h,jsob={A.app_ns+':transport':{'listen':{U:[]}},A.app_ns+':audit-log':{'log-entry':[]}}))
			A.global_root_id=D.inserted_primary_key[0];P=A.schema_path_to_real_table_name[_B+A.app_ns+':transport/listen/endpoint'];Q=A.metadata.tables[P]
			with A.engine.connect()as B:W=B.execute(Q.insert().values(pid=A.global_root_id,name=V,jsob={U:{G:V,'use-for':'native-interface','http':{'tcp-server-parameters':{'local-address':os.environ.get('SZTPD_DEFAULT_ADDR','127.0.0.1')}}}}))
	def _gen_tables(D,node,parent_table_name):
		R='ref_stats';G=parent_table_name;B=node
		if issubclass(type(B),yangson.schemanode.ListNode):
			C=[];C.append(sa.Column(_M,sa.Integer,primary_key=_E));N=D.schema_path_to_real_table_name[G];C.append(sa.Column(_I,sa.Integer,sa.ForeignKey(N+'.row_id'),index=_E,nullable=_C))
			if B.config==_E:
				if len(B.keys)>1:raise NotImplementedError('Only supports lists with at most one key.')
				E=B.get_child(*B.keys[0]);D.table_keys[B.data_path()]=E.name
				if type(E.type)==yangson.datatype.StringType:C.append(sa.Column(E.name,sa.String(250),nullable=_C))
				elif type(E.type)==yangson.datatype.Uint32Type:C.append(sa.Column(E.name,sa.Integer,nullable=_C))
				elif type(E.type)==yangson.datatype.IdentityrefType:C.append(sa.Column(E.name,sa.String(250),nullable=_C))
				elif type(E.type)==yangson.datatype.UnionType:C.append(sa.Column(E.name,sa.String(250),nullable=_C))
				else:raise Exception('Unsupported key node type: '+str(type(E.type)))
				if hasattr(E,'globally_unique'):C.append(sa.UniqueConstraint(E.name))
				else:C.append(sa.UniqueConstraint(E.name,_I))
				if B.user_ordered==_E:C.append(sa.Column(_H,sa.Integer,index=_E,nullable=_C))
				C.append(sa.Column(_O,jsob_type,nullable=_C))
			else:
				assert B.config==_C;assert hasattr(B,R)==_C
				for A in B.children:
					if issubclass(type(A),yangson.schemanode.LeafNode):
						if type(A.type)==yangson.datatype.StringType:
							if str(A.type)=='date-and-time(string)':C.append(sa.Column(A.name,sa.DateTime,index=_E,nullable=A.mandatory==_C or A.when!=_A))
							else:C.append(sa.Column(A.name,sa.String(250),index=_E,nullable=A.mandatory==_C or A.when!=_A))
						elif type(A.type)==yangson.datatype.Uint16Type:C.append(sa.Column(A.name,sa.SmallInteger,index=_E,nullable=A.mandatory==_C or A.when!=_A))
						elif type(A.type)==yangson.datatype.InstanceIdentifierType:C.append(sa.Column(A.name,sa.String(250),nullable=A.mandatory==_C or A.when!=_A))
						elif type(A.type)==yangson.datatype.LeafrefType:C.append(sa.Column(A.name,sa.String(250),nullable=A.mandatory==_C or A.when!=_A))
						elif type(A.type)==yangson.datatype.IdentityrefType:C.append(sa.Column(A.name,sa.String(250),nullable=A.mandatory==_C or A.when!=_A))
						elif type(A.type)==yangson.datatype.EnumerationType:C.append(sa.Column(A.name,sa.String(250),index=_E,nullable=A.mandatory==_C or A.when!=_A))
						elif type(A.type)==yangson.datatype.UnionType:
							J=0;S=_E
							for K in A.type.types:
								if issubclass(type(K),yangson.datatype.StringType):J+=1
								else:raise Exception('Unhandled union type: '+str(type(K)))
							if J==len(A.type.types):C.append(sa.Column(A.name,sa.String(250),index=_E,nullable=A.mandatory==_C or A.when!=_A))
							else:raise Exception('FIXME: not all union subtypes are stringafiable')
						else:raise Exception('Unhandled leaf type: '+str(type(A.type)))
					elif issubclass(type(A),yangson.schemanode.ChoiceNode):
						H=_E
						for I in A.children:
							assert type(I)==yangson.schemanode.CaseNode
							if len(I.children)>1:H=_C;break
							else:
								for O in I.children:
									if type(O)!=yangson.schemanode.LeafNode:H=_C;break
						if H==_E:C.append(sa.Column(A.name,sa.String(250),index=_E,nullable=A.mandatory==_C or A.when!=_A))
						else:C.append(sa.Column(A.name,jsob_type,nullable=A.mandatory==_C or A.when!=_A))
					elif issubclass(type(A),yangson.schemanode.AnydataNode):C.append(sa.Column(A.name,jsob_type,nullable=A.mandatory==_C or A.when!=_A))
					elif issubclass(type(A),yangson.schemanode.LeafListNode):C.append(sa.Column(A.name,jsob_type,nullable=A.mandatory==_C or A.when!=_A))
					elif issubclass(type(A),yangson.schemanode.ListNode):C.append(sa.Column(A.name,jsob_type,nullable=A.mandatory==_C or A.when!=_A))
					elif issubclass(type(A),yangson.schemanode.ContainerNode):C.append(sa.Column(A.name,jsob_type,nullable=A.mandatory==_C or A.when!=_A))
					elif issubclass(type(A),yangson.schemanode.NotificationNode):0
					else:raise Exception('Unhandled list child type: '+str(type(A)))
			P=re.sub('^/.*:','',B.data_path()).split(_B)
			if D.engine.dialect.name==_f:F=''
			else:F=D.db_schema+'.'
			for Q in P:F+=Q[0]
			while F in D.schema_path_to_real_table_name.values():F+='2'
			D.schema_path_to_real_table_name[B.data_path()]=F
			if D.db_schema is _A:L=sa.Table(F,D.metadata,*C)
			else:L=sa.Table(re.sub('^'+D.db_schema+'.','',F),D.metadata,*C)
			L.create();G=B.data_path()
		if B.config==_C and issubclass(type(B),yangson.schemanode.DataNode):
			M=B.data_path()
			if not any((M.startswith(A)for A in D.config_false_prefixes)):D.config_false_prefixes[M]=_E
		if B.config==_E and issubclass(type(B),yangson.schemanode.SequenceNode)and B.user_ordered==_E:D.config_true_obu_seq_nodes[B.data_path()]=_A
		if hasattr(B,R):D.ref_stat_collectors[B.data_path()]=_A
		if issubclass(type(B),yangson.schemanode.InternalNode):
			if not(type(B)==yangson.schemanode.ListNode and B.config==_C)and not type(B)==yangson.schemanode.RpcActionNode and not type(B)==yangson.schemanode.NotificationNode:
				for A in B.children:D._gen_tables(A,G)


# Copyright (c) 2021 Watsen Networks.  All Rights Reserved.

from __future__ import annotations
_Z='Unrecognized member: '
_Y='Query parameters are only supported for "list" and "leaf-list" nodes.'
_X='sleep'
_W=') is not a sibling of the target node ('
_V='Unrecognized schema path for "point" node: '
_U='The query parameter "point" is required when the "insert" parameter is "'
_T='Unrecognized "insert" query parameter value: '
_S='last'
_R='first'
_Q='" is not recognized.'
_P='The parameter "'
_O='The top-level node-identifier must be prefixed by a namespace followed by a colon.'
_N='Parent data node ('
_M='Input document must contain at least one top-level node.'
_L='Data node ('
_K='Unrecognized schema path: '
_J='Invalid data path: '
_I='Validation failed: '
_H='The "point" node ('
_G='after'
_F='before'
_E=') does not exist.'
_D=True
_C='insert'
_B='point'
_A='/'
import json,yangson,asyncio
from urllib.parse import quote
class NodeAlreadyExists(Exception):0
class NodeNotFound(Exception):0
class ParentNodeNotFound(Exception):0
class InvalidDataPath(Exception):0
class InvalidInputDocument(Exception):0
class UnrecognizedInputNode(Exception):0
class UnrecognizedQueryParameter(Exception):0
class MissingQueryParameter(Exception):0
class InvalidQueryParameter(Exception):0
class NonexistentSchemaNode(Exception):0
class ValidationFailed(Exception):0
class ValidationLayer:
	def __init__(A,dm,dal):
		A.dm=dm;A.dal=dal;C=asyncio.get_event_loop();D=A.dal.handle_get_config_request(_A);E=C.run_until_complete(D);A.inst=A.dm.from_raw(E)
		try:A.inst.validate()
		except yangson.exceptions.SchemaError as B:assert str(B).startswith('[/] missing-data: expected');assert str(B).endswith(":admin-accounts'")
	async def handle_get_config_request(C,data_path,query_dict):
		A=data_path;assert A!='';assert not(A!=_A and A[-1]==_A)
		try:E=C.dm.parse_resource_id(A)
		except yangson.exceptions.UnexpectedInput as B:raise InvalidDataPath(_J+str(B))
		except yangson.exceptions.NonexistentSchemaNode as B:raise NonexistentSchemaNode(_K+A)
		try:D=C.inst.goto(E)
		except yangson.exceptions.NonexistentInstance as B:raise NodeNotFound(_L+A+_E)
		if type(D)!=yangson.instance.ArrayEntry and isinstance(D.schema_node,yangson.schemanode.SequenceNode):raise InvalidDataPath("RFC 8040 doesn't acknowledge 'list' or 'list-list' nodes as resource targets.")
	async def handle_post_config_request(F,data_path,query_dict,request_body):
		f='=';e="' must contain one element.";d="Input 'list' node '";c="' not a 'list' node.";b="Input node '";S=None;H=data_path;C=request_body;B=query_dict
		if len(C)<1:raise InvalidInputDocument(_M)
		if len(C)>1:raise InvalidInputDocument('Input document must not have more than one top-level node.')
		try:T=F.dm.parse_resource_id(H)
		except yangson.exceptions.NonexistentSchemaNode as D:raise NonexistentSchemaNode('Unrecognized schema path for parent node: '+H)
		try:I=F.inst.goto(T)
		except yangson.exceptions.NonexistentInstance as D:raise ParentNodeNotFound(_N+H+_E)
		A=next(iter(C))
		if':'not in A:raise InvalidInputDocument(_O)
		U,G=A.split(':');M=I.schema_node;J=M.get_child(G,U)
		if J is S:raise UnrecognizedInputNode('Input document contains unrecognized top-level node.')
		if not M.ns is S:assert M.ns==J.ns
		if isinstance(J,yangson.schemanode.SequenceNode):
			if type(J)==yangson.schemanode.ListNode:
				V=J.keys[0];W=V[0]
				if type(C[A])!=list:raise InvalidInputDocument(b+G+c)
				if len(C[A])!=1:raise InvalidInputDocument(d+G+e)
				N=C[A][0];Q=N[W]
			elif type(J)==yangson.schemanode.LeafListNode:
				if type(C[A])!=list:raise InvalidInputDocument(b+G+c)
				if len(C[A])!=1:raise InvalidInputDocument(d+G+e)
				N=C[A][0];Q=C[A][0]
			else:raise AssertionError('Logic cannot reach this point')
			if H==_A:E=A;K=_A+E+f+quote(Q,safe='')
			else:E=G;K=H+_A+E+f+quote(Q,safe='')
			try:g=F.inst.raw_value();X=F.dm.parse_resource_id(K)
			except yangson.exceptions.NonexistentSchemaNode as D:raise NonexistentSchemaNode('Unrecognized schema path for insertion node: '+K)
			try:F.inst.goto(X)
			except yangson.exceptions.NonexistentInstance as D:pass
			else:raise NodeAlreadyExists('Child data node ('+K+') already exists.')
			for O in B:
				if O not in(_C,_B):raise UnrecognizedQueryParameter(_P+O+_Q)
				if O==_C:
					if B[_C]not in(_R,_F,_G,_S):raise InvalidQueryParameter(_T+B[_C])
					if B[_C]in(_F,_G):
						if _B not in B:raise MissingQueryParameter(_U+insert+'"')
				if O==_B:
					try:Y=F.dm.parse_resource_id(B[_B])
					except yangson.exceptions.NonexistentSchemaNode as D:raise NonexistentSchemaNode(_V+B[_B])
					try:F.inst.goto(Y)
					except yangson.exceptions.NonexistentInstance as D:raise InvalidQueryParameter(_H+B[_B]+_E)
					if H!=B[_B].rsplit(_A,1)[0]:raise InvalidQueryParameter(_H+B[_B]+_W+K+').')
			try:L=I[E]
			except yangson.exceptions.NonexistentInstance:L=I.put_member(E,yangson.instvalue.ArrayValue([]))
			assert isinstance(L.schema_node,yangson.schemanode.SequenceNode)
			if len(L.value)==0:
				try:Z=L.update([N],raw=_D)
				except yangson.exceptions.RawMemberError as D:raise UnrecognizedInputNode('Incompatable node data. '+str(D))
				P=Z.up()
			else:a=L[-1];P=a.insert_after(N,raw=_D).up()
		else:
			if len(B)>0 and len(B)==1 and _X not in B:raise UnrecognizedQueryParameter(_Y)
			if H==_A:E=A
			else:E=G
			if E in I:raise NodeAlreadyExists('Node "'+E+'" already exists.')
			try:
				if M.ns==S:P=I.put_member(A,C[A],raw=_D).up()
				else:P=I.put_member(G,C[A],raw=_D).up()
			except yangson.exceptions.RawMemberError as D:raise UnrecognizedInputNode(_Z+str(D))
		R=P.top()
		try:R.validate()
		except Exception as D:raise ValidationFailed(_I+str(D))
		F.inst2=R
	async def handle_put_config_request(C,data_path,query_dict,request_body):
		E=request_body;D=data_path;A=query_dict;assert D!='';assert not(D!=_A and D[-1]==_A)
		if len(E)<1:raise InvalidInputDocument(_M)
		I=next(iter(E))
		if':'not in I:raise InvalidInputDocument(_O)
		try:L=C.dm.parse_resource_id(D)
		except yangson.exceptions.UnexpectedInput as B:raise InvalidDataPath(_J+str(B))
		except yangson.exceptions.NonexistentSchemaNode as B:raise NonexistentSchemaNode(_K+D)
		try:F=C.inst.goto(L)
		except yangson.exceptions.NonexistentInstance as B:
			O=C.inst.raw_value();G=D.rsplit(_A,1)[0]
			if G=='':G=_A
			M=C.dm.parse_resource_id(G)
			try:F=C.inst.goto(M)
			except yangson.exceptions.NonexistentInstance as B:raise ParentNodeNotFound(_N+G+') does not exist. '+str(B))
			await C.handle_post_config_request(G,A,E);return
		if isinstance(F.schema_node,yangson.schemanode.SequenceNode):
			for H in A:
				if H not in(_C,_B):raise UnrecognizedQueryParameter(_P+H+_Q)
				if H==_C:
					if A[_C]not in(_R,_F,_G,_S):raise InvalidQueryParameter(_T+A[_C])
					if A[_C]in(_F,_G):
						if _B not in A:raise MissingQueryParameter(_U+insert+'"')
				if H==_B:
					try:N=C.dm.parse_resource_id(A[_B])
					except yangson.exceptions.NonexistentSchemaNode as B:raise NonexistentSchemaNode(_V+A[_B])
					try:C.inst.goto(N)
					except yangson.exceptions.NonexistentInstance as B:raise InvalidQueryParameter(_H+A[_B]+_E)
					if D.rsplit(_A,1)[0]!=A[_B].rsplit(_A,1)[0]:raise InvalidQueryParameter(_H+A[_B]+_W+D+').')
		elif len(A)>0 and len(A)==1 and _X not in A:raise UnrecognizedQueryParameter(_Y)
		try:
			if D==_A:J=F.update(E,raw=_D)
			elif isinstance(F.schema_node,yangson.schemanode.SequenceNode):J=F.update(E[I][0],raw=_D)
			else:J=F.update(E[I],raw=_D)
		except yangson.exceptions.RawMemberError as B:raise UnrecognizedInputNode(_Z+str(B))
		except Exception as B:raise NotImplementedError(str(type(B))+' = '+str(B))
		K=J.top()
		try:K.validate()
		except Exception as B:raise ValidationFailed(_I+str(B))
		C.inst2=K
	async def handle_delete_config_request(C,data_path):
		B=data_path;assert B!=''
		if B==_A:E=RootNode(ObjectValue({}),C.inst.schema_node,datetime.now())
		else:
			try:J=C.dm.parse_resource_id(B)
			except yangson.exceptions.NonexistentSchemaNode as F:raise NonexistentSchemaNode('Unrecognized schema path for data node: '+B)
			try:D=C.inst.goto(J)
			except yangson.exceptions.NonexistentInstance as F:raise NodeNotFound(_L+B+_E)
			H=D.up()
			if type(D)==yangson.instance.ArrayEntry:
				A=H.delete_item(D.index)
				if len(A.value)==0:
					G=A.up()
					if isinstance(G.schema_node,yangson.schemanode.SequenceNode):I=G.delete_item(A.index);raise NotImplementedError('tested? list inside a list...')
					else:I=G.delete_item(A.name)
					A=I
			else:A=H.delete_item(D.name)
			E=A.top()
		try:E.validate()
		except Exception as F:raise ValidationFailed(_I+str(F))
		C.inst2=E


# Copyright (c) 2021 Watsen Networks.  All Rights Reserved.

from __future__ import annotations
_C='/ds/ietf-datastores:operational'
_B='/ds/ietf-datastores:running'
_A=None
import os,ssl,json,base64,pyasn1,asyncio,yangson,datetime,tempfile,basicauth
from .  import utils
from aiohttp import web
from .handler import RouteHandler
from pyasn1.type import univ
from pyasn1_modules import rfc3447
from pyasn1_modules import rfc5280
from pyasn1_modules import rfc5652
from pyasn1_modules import rfc5915
from pyasn1.codec.der.decoder import decode as der_decoder
from pyasn1.codec.der.encoder import encode as der_encoder
async def set_server_header(request,response):response.headers['Server']='<redacted>'
class RestconfServer:
	root='/restconf';prefix_running=root+_B;prefix_operational=root+_C;prefix_operations=root+'/operations';len_prefix_running=len(prefix_running);len_prefix_operational=len(prefix_operational);len_prefix_operations=len(prefix_operations)
	def __init__(A,loop,dal,endpoint_config,view_handler,facade_yl=_A):
		AC='client-certs';AB='local-truststore-reference';AA='ca-certs';A9='client-authentication';A8='\n-----END CERTIFICATE-----\n';A7='-----BEGIN CERTIFICATE-----\n';A6='cert-data';A5='private-key-format';A4=':keystore/asymmetric-keys/asymmetric-key=';A3='reference';A2='server-identity';A1='local-port';A0='http';j='ASCII';i=':asymmetric-key';h='tcp-server-parameters';S='certificate';R='tls-server-parameters';Q='/ds/ietf-datastores:running{tail:.*}';P='/';G=dal;E='https';C=endpoint_config;B=view_handler;A.len_prefix_running=len(A.root+_B);A.len_prefix_operational=len(A.root+_C);A.loop=loop;A.dal=G;A.name=C['name'];A.view_handler=B;A.app=web.Application();A.app.on_response_prepare.append(set_server_header);A.app.router.add_get('/.well-known/host-meta',A.handle_get_host_meta);A.app.router.add_get(A.root,B.handle_get_restconf_root);A.app.router.add_get(A.root+P,B.handle_get_restconf_root);A.app.router.add_get(A.root+'/yang-library-version',B.handle_get_yang_library_version);A.app.router.add_get(A.root+'/ds/ietf-datastores:operational{tail:.*}',B.handle_get_opstate_request);A.app.router.add_get(A.root+Q,B.handle_get_config_request);A.app.router.add_put(A.root+Q,B.handle_put_config_request);A.app.router.add_post(A.root+Q,B.handle_post_config_request);A.app.router.add_delete(A.root+Q,B.handle_delete_config_request);A.app.router.add_post(A.root+'/ds/ietf-datastores:operational/{tail:.*}',B.handle_action_request);A.app.router.add_post(A.root+'/operations/{tail:.*}',B.handle_rpc_request)
		if A0 in C:F=A0
		else:assert E in C;F=E
		A.local_address=C[F][h]['local-address'];A.local_port=os.environ.get('SZTPD_DEFAULT_PORT',8080)
		if A1 in C[F][h]:A.local_port=C[F][h][A1]
		D=_A
		if F==E:
			T=C[E][R][A2][S][A3]['asymmetric-key'];J=A.dal.handle_get_config_request(P+A.dal.app_ns+A4+T);K=A.loop.run_until_complete(J);L=K[A.dal.app_ns+i][0]['cleartext-private-key'];U=base64.b64decode(L)
			if K[A.dal.app_ns+i][0][A5]=='ietf-crypto-types:ec-private-key-format':M,k=der_decoder(U,asn1Spec=rfc5915.ECPrivateKey());l=der_encoder(M);V=base64.b64encode(l).decode(j);assert L==V;W='-----BEGIN EC PRIVATE KEY-----\n'+V+'\n-----END EC PRIVATE KEY-----\n'
			elif K[A.dal.app_ns+i][0][A5]=='ietf-crypto-types:rsa-private-key-format':M,k=der_decoder(U,asn1Spec=rfc3447.RSAPrivateKey());m=der_encoder(M);X=base64.b64encode(m).decode(j);assert L==X;W='-----BEGIN RSA PRIVATE KEY-----\n'+X+'\n-----END RSA PRIVATE KEY-----\n'
			else:raise NotImplementedError('this line can never be reached')
			n=C[E][R][A2][S][A3][S];J=A.dal.handle_get_config_request(P+A.dal.app_ns+A4+T+'/certificates/certificate='+n);o=A.loop.run_until_complete(J);p=o[A.dal.app_ns+':certificate'][0][A6];q=base64.b64decode(p);r,Y=der_decoder(q,asn1Spec=rfc5652.ContentInfo());s=r.getComponentByName('content');t,Y=der_decoder(s,asn1Spec=rfc5652.SignedData());Z=t.getComponentByName('certificates');H=''
			for u in range(len(Z)):
				a=Z[u][0]
				for b in a['tbsCertificate']['extensions']:
					if b['extnID']==rfc5280.id_ce_basicConstraints:v,Y=der_decoder(b['extnValue'],asn1Spec=rfc5280.BasicConstraints())
				w=der_encoder(a);c=base64.b64encode(w).decode(j)
				if v['cA']==False:H=A7+c+A8+H
				else:H+=A7+c+A8
			D=ssl.create_default_context(ssl.Purpose.CLIENT_AUTH);D.verify_mode=ssl.CERT_OPTIONAL
			with tempfile.TemporaryDirectory()as d:
				e=d+'key.pem';f=d+'certs.pem'
				with open(e,'w')as x:x.write(W)
				with open(f,'w')as y:y.write(H)
				D.load_cert_chain(f,e)
			if A9 in C[E][R]:
				I=C[E][R][A9]
				def g(truststore_ref):
					C=G.handle_get_config_request(P+G.app_ns+':truststore/certificate-bags/certificate-bag='+truststore_ref);D=A.loop.run_until_complete(C);B=[]
					for E in D[G.app_ns+':certificate-bag'][0][S]:F=base64.b64decode(E[A6]);H,I=der_decoder(F,asn1Spec=rfc5652.ContentInfo());assert not I;B+=utils.degenerate_cms_obj_to_ders(H)
					return B
				N=[]
				if AA in I:O=I[AA][AB];N+=g(O)
				if AC in I:O=I[AC][AB];N+=g(O)
				z=utils.der_dict_to_multipart_pem({'CERTIFICATE':N});D.load_verify_locations(cadata=z)
		if F==E:assert not D is _A
		else:assert D is _A
		A.runner=web.AppRunner(A.app);A.loop.run_until_complete(A.runner.setup());A.site=web.TCPSite(A.runner,host=A.local_address,port=A.local_port,ssl_context=D,reuse_port=True);A.loop.run_until_complete(A.site.start())
	async def handle_get_host_meta(B,request):A=web.Response();A.content_type='application/xrd+xml';A.text='<XRD xmlns="http://docs.oasis-open.org/ns/xri/xrd-1.0">\n  <Link rel="restconf" href="/restconf"/>\n</XRD>';return A


# Copyright (c) 2021 Watsen Networks.  All Rights Reserved.

from __future__ import annotations
_Ag='Returning an RPC-error provided by callback (NOTE: RPC-error != exception, hence a normal exit).'
_Af='Unrecognized error-tag: '
_Ae='partial-operation'
_Ad='operation-failed'
_Ac='rollback-failed'
_Ab='data-exists'
_Aa='resource-denied'
_AZ='lock-denied'
_AY='in-use'
_AX='unknown-namespace'
_AW='bad-element'
_AV='unknown-attribute'
_AU='bad-attribute'
_AT='missing-attribute'
_AS='too-big'
_AR='exception-thrown'
_AQ='functions'
_AP='callback-details'
_AO='from-device'
_AN='identity-certificate'
_AM='source-ip-address'
_AL='mode-0 == no-sn'
_AK='"ietf-sztp-bootstrap-server:input" is missing.'
_AJ='/ietf-sztp-bootstrap-server:report-progress'
_AI='Resource does not exist.'
_AH='Requested resource does not exist.'
_AG=':log-entry'
_AF='/devices/device='
_AE=':devices/device='
_AD='2021-02-04'
_AC='2019-04-30'
_AB='urn:ietf:params:xml:ns:yang:ietf-yang-types'
_AA='ietf-yang-types'
_A9='module'
_A8='module-set-id'
_A7='ietf-yang-library:modules-state'
_A6='application/yang-data+xml'
_A5='webhooks'
_A4='callout-type'
_A3='passed-input'
_A2='ssl_object'
_A1='access-denied'
_A0='/ietf-sztp-bootstrap-server:get-bootstrapping-data'
_z='Parent node does not exist.'
_y='Resource can not be modified.'
_x='2013-07-15'
_w='webhook'
_v='exited-normally'
_u='opaque'
_t='function'
_s='plugin'
_r='serial-number'
_q='rpc-supported'
_p='data-missing'
_o='Unable to parse "input" document: '
_n='import'
_m='application/yang-data+json'
_l='operation-not-supported'
_k=':device'
_j='Content-Type'
_i='1'
_h='malformed-message'
_g=False
_f=':tenants/tenant='
_e='x'
_d='implement'
_c='callback-results'
_b='callback'
_a='invalid-value'
_Z='unknown-element'
_Y=True
_X='application'
_W='path'
_V='method'
_U='source-ip'
_T='timestamp'
_S='0'
_R='conformance-type'
_Q='namespace'
_P='revision'
_O='ietf-sztp-bootstrap-server:input'
_N='error-tag'
_M='error'
_L='protocol'
_K='text/plain'
_J='ietf-restconf:errors'
_I='name'
_H=':dynamic-callout'
_G='+'
_F='return-code'
_E='dynamic-callout'
_D='error-returned'
_C=None
_B='event-details'
_A='/'
import os,json,base64,pprint,asyncio,aiohttp,yangson,datetime,basicauth,urllib.parse,pkg_resources
from .  import yl
from .  import dal
from .  import utils
from aiohttp import web
from .native import Read
from pyasn1.type import univ
from .dal import DataAccessLayer
from .rcsvr import RestconfServer
from .handler import RouteHandler
from pyasn1_modules import rfc5652
from passlib.hash import sha256_crypt
from pyasn1.codec.der.encoder import encode as encode_der
from pyasn1.codec.der.decoder import decode as der_decoder
from certvalidator import CertificateValidator,ValidationContext,PathBuildingError
from cryptography.hazmat.backends import default_backend
from cryptography import x509
class RFC8572ViewHandler(RouteHandler):
	len_prefix_running=len(RestconfServer.root+'/ds/ietf-datastores:running');len_prefix_operational=len(RestconfServer.root+'/ds/ietf-datastores:operational');len_prefix_operations=len(RestconfServer.root+'/operations');id_ct_sztpConveyedInfoXML=rfc5652._buildOid(1,2,840,113549,1,9,16,1,42);id_ct_sztpConveyedInfoJSON=rfc5652._buildOid(1,2,840,113549,1,9,16,1,43);supported_media_types=_m,_A6;yl4errors={_A7:{_A8:'TBD',_A9:[{_I:_AA,_P:_x,_Q:_AB,_R:_n},{_I:'ietf-restconf',_P:'2017-01-26',_Q:'urn:ietf:params:xml:ns:yang:ietf-restconf',_R:_d},{_I:'ietf-netconf-acm',_P:'2018-02-14',_Q:'urn:ietf:params:xml:ns:yang:ietf-netconf-acm',_R:_n},{_I:'ietf-sztp-bootstrap-server',_P:_AC,_Q:'urn:ietf:params:xml:ns:yang:ietf-sztp-bootstrap-server',_R:_d},{_I:'ietf-yang-structure-ext',_P:'2020-06-22',_Q:'urn:ietf:params:xml:ns:yang:ietf-yang-structure-ext',_R:_d},{_I:'ietf-sztp-csr',_P:_AD,_Q:'urn:ietf:params:xml:ns:yang:ietf-sztp-csr',_R:_d},{_I:'ietf-crypto-types',_P:_AD,_Q:'urn:ietf:params:xml:ns:yang:ietf-crypto-types',_R:_d}]}};yl4conveyedinfo={_A7:{_A8:'TBD',_A9:[{_I:_AA,_P:_x,_Q:_AB,_R:_n},{_I:'ietf-inet-types',_P:_x,_Q:'urn:ietf:params:xml:ns:yang:ietf-inet-types',_R:_n},{_I:'ietf-sztp-conveyed-info',_P:_AC,_Q:'urn:ietf:params:xml:ns:yang:ietf-sztp-conveyed-info',_R:_d}]}}
	def __init__(A,dal,mode,yl,nvh):D='sztpd';A.dal=dal;A.mode=mode;A.nvh=nvh;B=pkg_resources.resource_filename(D,'yang');A.dm=yangson.DataModel(json.dumps(yl),[B]);A.dm4conveyedinfo=yangson.DataModel(json.dumps(A.yl4conveyedinfo),[B]);C=pkg_resources.resource_filename(D,'yang4errors');A.dm4errors=yangson.DataModel(json.dumps(A.yl4errors),[C,B])
	async def _insert_bootstrapping_log_entry(A,device_id,bootstrapping_log_entry):
		E='/bootstrapping-log';B=device_id
		if A.mode==_S:C=_A+A.dal.app_ns+':device/bootstrapping-log'
		elif A.mode==_i:C=_A+A.dal.app_ns+_AE+B[0]+E
		elif A.mode==_e:C=_A+A.dal.app_ns+_f+B[1]+_AF+B[0]+E
		D={};D[A.dal.app_ns+_AG]=bootstrapping_log_entry;await A.dal.handle_post_opstate_request(C,D)
	async def _insert_audit_log_entry(A,tenant_name,audit_log_entry):
		B=tenant_name
		if A.mode==_S or A.mode==_i or B==_C:C=_A+A.dal.app_ns+':audit-log'
		elif A.mode==_e:C=_A+A.dal.app_ns+_f+B+'/audit-log'
		D={};D[A.dal.app_ns+_AG]=audit_log_entry;await A.dal.handle_post_opstate_request(C,D)
	async def handle_get_restconf_root(D,request):
		C=request;J=_A;F=await D._check_auth(C,J)
		if type(F)is web.Response:A=F;return A
		else:H=F
		B={};B[_T]=datetime.datetime.utcnow();B[_U]=C.remote;B[_V]=C.method;B[_W]=C.path;E,K=utils.check_http_headers(C,D.supported_media_types,accept_required=_Y)
		if type(E)is web.Response:A=E;L=K;B[_F]=A.status;B[_D]=L;await D._insert_bootstrapping_log_entry(H,B);return A
		else:assert type(E)==str;G=E;assert G!=_K;I=utils.Encoding[G.rsplit(_G,1)[1]]
		A=web.Response(status=200);A.content_type=G
		if I==utils.Encoding.json:A.text='{\n    "ietf-restconf:restconf" : {\n        "data" : {},\n        "operations" : {},\n        "yang-library-version" : "2019-01-04"\n    }\n}\n'
		else:assert I==utils.Encoding.xml;A.text='<restconf xmlns="urn:ietf:params:xml:ns:yang:ietf-restconf">\n    <data/>\n    <operations/>\n    <yang-library-version>2016-06-21</yang-library-version>\n</restconf>\n'
		B[_F]=A.status;await D._insert_bootstrapping_log_entry(H,B);return A
	async def handle_get_yang_library_version(D,request):
		C=request;J=_A;F=await D._check_auth(C,J)
		if type(F)is web.Response:A=F;return A
		else:H=F
		B={};B[_T]=datetime.datetime.utcnow();B[_U]=C.remote;B[_V]=C.method;B[_W]=C.path;E,K=utils.check_http_headers(C,D.supported_media_types,accept_required=_Y)
		if type(E)is web.Response:A=E;L=K;B[_F]=A.status;B[_D]=L;await D._insert_bootstrapping_log_entry(H,B);return A
		else:assert type(E)==str;G=E;assert G!=_K;I=utils.Encoding[G.rsplit(_G,1)[1]]
		A=web.Response(status=200);A.content_type=G
		if I==utils.Encoding.json:A.text='{\n  "ietf-restconf:yang-library-version" : "2019-01-04"\n}'
		else:assert I==utils.Encoding.xml;A.text='<yang-library-version xmlns="urn:ietf:params:xml:ns:yang:ietf-restconf">2019-01-04</yang-library-version>'
		B[_F]=A.status;await D._insert_bootstrapping_log_entry(H,B);return A
	async def handle_get_opstate_request(C,request):
		D=request;E=D.path[C.len_prefix_operational:];E=_A;G=await C._check_auth(D,E)
		if type(G)is web.Response:A=G;return A
		else:I=G
		B={};B[_T]=datetime.datetime.utcnow();B[_U]=D.remote;B[_V]=D.method;B[_W]=D.path;F,L=utils.check_http_headers(D,C.supported_media_types,accept_required=_Y)
		if type(F)is web.Response:A=F;M=L;B[_F]=A.status;B[_D]=M;await C._insert_bootstrapping_log_entry(I,B);return A
		else:assert type(F)==str;H=F;assert H!=_K;J=utils.Encoding[H.rsplit(_G,1)[1]]
		if E=='/ietf-yang-library:yang-library'or E==_A or E=='':A=web.Response(status=200);A.content_type=_m;A.text=getattr(yl,'sbi_rfc8572')()
		else:A=web.Response(status=404);A.content_type=H;J=utils.Encoding[A.content_type.rsplit(_G,1)[1]];K=utils.gen_rc_errors(_L,_Z,error_message=_AH);N=C.dm4errors.get_schema_node(_A);A.text=utils.obj_to_encoded_str(K,J,C.dm4errors,N);B[_D]=K
		B[_F]=A.status;await C._insert_bootstrapping_log_entry(I,B);return A
	async def handle_get_config_request(C,request):
		D=request;F=D.path[C.len_prefix_running:];G=await C._check_auth(D,F)
		if type(G)is web.Response:A=G;return A
		else:I=G
		B={};B[_T]=datetime.datetime.utcnow();B[_U]=D.remote;B[_V]=D.method;B[_W]=D.path;E,L=utils.check_http_headers(D,C.supported_media_types,accept_required=_Y)
		if type(E)is web.Response:A=E;M=L;B[_F]=A.status;B[_D]=M;await C._insert_bootstrapping_log_entry(I,B);return A
		else:assert type(E)==str;H=E;assert H!=_K;J=utils.Encoding[H.rsplit(_G,1)[1]]
		if F==_A or F=='':A=web.Response(status=204)
		else:A=web.Response(status=404);A.content_type=H;J=utils.Encoding[A.content_type.rsplit(_G,1)[1]];K=utils.gen_rc_errors(_L,_Z,error_message=_AH);N=C.dm4errors.get_schema_node(_A);A.text=utils.obj_to_encoded_str(K,J,C.dm4errors,N);B[_D]=K
		B[_F]=A.status;await C._insert_bootstrapping_log_entry(I,B);return A
	async def handle_post_config_request(C,request):
		D=request;F=D.path[C.len_prefix_running:];G=await C._check_auth(D,F)
		if type(G)is web.Response:A=G;return A
		else:J=G
		B={};B[_T]=datetime.datetime.utcnow();B[_U]=D.remote;B[_V]=D.method;B[_W]=D.path;E,L=utils.check_http_headers(D,C.supported_media_types,accept_required=_g)
		if type(E)is web.Response:A=E;M=L;B[_F]=A.status;B[_D]=M;await C._insert_bootstrapping_log_entry(J,B);return A
		else:assert type(E)==str;H=E;assert H!=_K;K=utils.Encoding[H.rsplit(_G,1)[1]]
		if F==_A or F=='':A=web.Response(status=400);I=utils.gen_rc_errors(_X,_a,error_message=_y)
		else:A=web.Response(status=404);I=utils.gen_rc_errors(_L,_Z,error_message=_z)
		A.content_type=H;K=utils.Encoding[A.content_type.rsplit(_G,1)[1]];N=C.dm4errors.get_schema_node(_A);A.text=utils.obj_to_encoded_str(I,K,C.dm4errors,N);B[_F]=A.status;B[_D]=I;await C._insert_bootstrapping_log_entry(J,B);return A
	async def handle_put_config_request(C,request):
		D=request;F=D.path[C.len_prefix_running:];G=await C._check_auth(D,F)
		if type(G)is web.Response:A=G;return A
		else:J=G
		B={};B[_T]=datetime.datetime.utcnow();B[_U]=D.remote;B[_V]=D.method;B[_W]=D.path;E,L=utils.check_http_headers(D,C.supported_media_types,accept_required=_g)
		if type(E)is web.Response:A=E;M=L;B[_F]=A.status;B[_D]=M;await C._insert_bootstrapping_log_entry(J,B);return A
		else:assert type(E)==str;H=E;assert H!=_K;K=utils.Encoding[H.rsplit(_G,1)[1]]
		if F==_A or F=='':A=web.Response(status=400);I=utils.gen_rc_errors(_X,_a,error_message=_y)
		else:A=web.Response(status=404);I=utils.gen_rc_errors(_L,_Z,error_message=_z)
		A.content_type=H;K=utils.Encoding[A.content_type.rsplit(_G,1)[1]];N=C.dm4errors.get_schema_node(_A);A.text=utils.obj_to_encoded_str(I,K,C.dm4errors,N);B[_F]=A.status;B[_D]=I;await C._insert_bootstrapping_log_entry(J,B);return A
	async def handle_delete_config_request(C,request):
		D=request;G=D.path[C.len_prefix_running:];H=await C._check_auth(D,G)
		if type(H)is web.Response:A=H;return A
		else:L=H
		B={};B[_T]=datetime.datetime.utcnow();B[_U]=D.remote;B[_V]=D.method;B[_W]=D.path;E,M=utils.check_http_headers(D,C.supported_media_types,accept_required=_g)
		if type(E)is web.Response:A=E;N=M;B[_F]=A.status;B[_D]=N;await C._insert_bootstrapping_log_entry(L,B);return A
		else:
			assert type(E)==str;I=E
			if I==_K:J=_C
			else:J=utils.Encoding[I.rsplit(_G,1)[1]]
		if G==_A or G=='':A=web.Response(status=400);F=_y;K=utils.gen_rc_errors(_X,_a,error_message=F)
		else:A=web.Response(status=404);F=_z;K=utils.gen_rc_errors(_L,_Z,error_message=F)
		A.content_type=I
		if J is _C:A.text=F
		else:O=C.dm4errors.get_schema_node(_A);A.text=utils.obj_to_encoded_str(K,J,C.dm4errors,O)
		B[_F]=A.status;B[_D]=K;await C._insert_bootstrapping_log_entry(L,B);return A
	async def handle_action_request(C,request):
		D=request;F=D.path[C.len_prefix_operational:];G=await C._check_auth(D,F)
		if type(G)is web.Response:A=G;return A
		else:J=G
		B={};B[_T]=datetime.datetime.utcnow();B[_U]=D.remote;B[_V]=D.method;B[_W]=D.path;E,L=utils.check_http_headers(D,C.supported_media_types,accept_required=_g)
		if type(E)is web.Response:A=E;M=L;B[_F]=A.status;B[_D]=M;await C._insert_bootstrapping_log_entry(J,B);return A
		else:assert type(E)==str;H=E;assert H!=_K;K=utils.Encoding[H.rsplit(_G,1)[1]]
		if F==_A or F=='':A=web.Response(status=400);I=utils.gen_rc_errors(_X,_a,error_message='Resource does not support action.')
		else:A=web.Response(status=404);I=utils.gen_rc_errors(_L,_Z,error_message=_AI)
		A.content_type=H;K=utils.Encoding[A.content_type.rsplit(_G,1)[1]];N=C.dm4errors.get_schema_node(_A);A.text=utils.obj_to_encoded_str(I,K,C.dm4errors,N);B[_F]=A.status;B[_D]=I;await C._insert_bootstrapping_log_entry(J,B);return A
	async def handle_rpc_request(C,request):
		P='sleep';D=request;F=D.path[C.len_prefix_operations:];J=await C._check_auth(D,F)
		if type(J)is web.Response:A=J;return A
		else:E=J
		B={};B[_T]=datetime.datetime.utcnow();B[_U]=D.remote;B[_V]=D.method;B[_W]=D.path;H,M=utils.check_http_headers(D,C.supported_media_types,accept_required=_g)
		if type(H)is web.Response:A=H;N=M;B[_F]=A.status;B[_D]=N;await C._insert_bootstrapping_log_entry(E,B);return A
		else:
			assert type(H)==str;K=H
			if K==_K:L=_C
			else:L=utils.Encoding[K.rsplit(_G,1)[1]]
		if F==_A0:
			async with C.nvh.fifolock(Read):
				if os.environ.get('SZTPD_MODE')and P in D.query:await asyncio.sleep(int(D.query[P]))
				A=await C._handle_get_bootstrapping_data_rpc(E,D,B);B[_F]=A.status;await C._insert_bootstrapping_log_entry(E,B);return A
		elif F==_AJ:
			try:A=await C._handle_report_progress_rpc(E,D,B)
			except NotImplementedError as Q:raise NotImplementedError('is this ever called?')
			B[_F]=A.status;await C._insert_bootstrapping_log_entry(E,B);return A
		elif F==_A or F=='':A=web.Response(status=400);G=_AI;I=utils.gen_rc_errors(_X,_a,error_message=G)
		else:A=web.Response(status=404);G='Unrecognized RPC.';I=utils.gen_rc_errors(_L,_Z,error_message=G)
		A.content_type=K
		if A.content_type==_K:A.text=G
		else:I=utils.gen_rc_errors(_L,_a,error_message=G);O=C.dm4errors.get_schema_node(_A);A.text=utils.obj_to_encoded_str(I,L,C.dm4errors,O)
		B[_F]=A.status;B[_D]=I;await C._insert_bootstrapping_log_entry(E,B);return A
	async def _check_auth(A,request,data_path):
		z='num-times-accessed';y='local-truststore-reference';x=':device-type';w='identity-certificates';v='activation-code';u='" not found for any tenant.';t='Device "';s='X-Client-Cert';e='verification';d='device-type';T='sbi-access-stats';P='lifecycle-statistics';J='comment';I='failure';H='outcome';C=request
		def F(request,supported_media_types):
			E='Accept';D=supported_media_types;C=request;B=web.Response(status=401)
			if E in C.headers and any((C.headers[E]==A for A in D)):B.content_type=C.headers[E]
			elif _j in C.headers and any((C.headers[_j]==A for A in D)):B.content_type=C.headers[_j]
			else:B.content_type=_K
			if B.content_type!=_K:F=utils.Encoding[B.content_type.rsplit(_G,1)[1]];G=utils.gen_rc_errors(_L,_A1);H=A.dm4errors.get_schema_node(_A);B.text=utils.obj_to_encoded_str(G,F,A.dm4errors,H)
			return B
		B={};B[_T]=datetime.datetime.utcnow();B[_U]=C.remote;B['source-proxies']=list(C.forwarded);B['host']=C.host;B[_V]=C.method;B[_W]=C.path;K=set();M=_C;N=C.transport.get_extra_info('peercert')
		if N is not _C:O=N['subject'][-1][0][1];K.add(O)
		elif C.headers.get(s)!=_C:f=C.headers.get(s);U=bytes(urllib.parse.unquote(f),'utf-8');M=x509.load_pem_x509_certificate(U,default_backend());g=M.subject;O=g.get_attributes_for_oid(x509.ObjectIdentifier('2.5.4.5'))[0].value;K.add(O)
		Q=_C;V=_C;R=C.headers.get('AUTHORIZATION')
		if R!=_C:Q,V=basicauth.decode(R);K.add(Q)
		if len(K)==0:B[H]=I;B[J]='Device provided no identification credentials.';await A._insert_audit_log_entry(_C,B);return F(C,A.supported_media_types)
		if len(K)!=1:B[H]=I;B[J]='Device provided mismatched authentication credentials ('+O+' != '+Q+').';await A._insert_audit_log_entry(_C,B);return F(C,A.supported_media_types)
		E=K.pop();D=_C
		if A.mode==_S:L=_A+A.dal.app_ns+_k
		elif A.mode==_i:L=_A+A.dal.app_ns+_AE+E
		if A.mode!=_e:
			try:D=await A.dal.handle_get_opstate_request(L)
			except dal.NodeNotFound as W:B[H]=I;B[J]=t+E+u;await A._insert_audit_log_entry(_C,B);return F(C,A.supported_media_types)
			G=_C
		else:
			try:G=await A.dal.get_tenant_name_for_global_key(_A+A.dal.app_ns+':tenants/tenant/devices/device',E)
			except dal.NodeNotFound as W:B[H]=I;B[J]=t+E+u;await A._insert_audit_log_entry(_C,B);return F(C,A.supported_media_types)
			L=_A+A.dal.app_ns+_f+G+_AF+E;D=await A.dal.handle_get_opstate_request(L)
		assert D!=_C;assert A.dal.app_ns+_k in D;D=D[A.dal.app_ns+_k]
		if A.mode!=_S:D=D[0]
		if v in D:
			if R==_C:B[H]=I;B[J]='Activation code required but none passed for serial number '+E;await A._insert_audit_log_entry(G,B);return F(C,A.supported_media_types)
			X=D[v];assert X.startswith('$5$')
			if not sha256_crypt.verify(V,X):B[H]=I;B[J]='Activation code mismatch for serial number '+E;await A._insert_audit_log_entry(G,B);return F(C,A.supported_media_types)
		else:0
		assert d in D;h=_A+A.dal.app_ns+':device-types/device-type='+D[d];Y=await A.dal.handle_get_opstate_request(h)
		if w in Y[A.dal.app_ns+x][0]:
			if N is _C and M is _C:B[H]=I;B[J]='Client cert required but none passed for serial number '+E;await A._insert_audit_log_entry(G,B);return F(C,A.supported_media_types)
			if N:Z=C.transport.get_extra_info(_A2);assert Z is not _C;a=Z.getpeercert(_Y)
			else:assert M is not _C;a=U
			S=Y[A.dal.app_ns+x][0][w];assert e in S;assert y in S[e];b=S[e][y];i=_A+A.dal.app_ns+':truststore/certificate-bags/certificate-bag='+b['certificate-bag']+'/certificate='+b['certificate'];j=await A.dal.handle_get_config_request(i);k=j[A.dal.app_ns+':certificate'][0]['cert-data'];l=base64.b64decode(k);m,n=der_decoder(l,asn1Spec=rfc5652.ContentInfo());assert not n;o=utils.degenerate_cms_obj_to_ders(m);p=ValidationContext(trust_roots=o);q=CertificateValidator(a,validation_context=p)
			try:q._validate_path()
			except PathBuildingError as W:B[H]=I;B[J]="Client cert for serial number '"+E+"' does not validate using trust anchors specified by device-type '"+D[d]+"'";await A._insert_audit_log_entry(G,B);return F(C,A.supported_media_types)
		B[H]='success';await A._insert_audit_log_entry(G,B);r=L+'/lifecycle-statistics';c=datetime.datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%SZ')
		if D[P][T][z]==0:D[P][T]['first-accessed']=c
		D[P][T]['last-accessed']=c;D[P][T][z]+=1;await A.dal.handle_put_opstate_request(r,D[P]);return E,G
	async def _handle_get_bootstrapping_data_rpc(A,device_id,request,bootstrapping_log_entry):
		Aj='ietf-sztp-bootstrap-server:output';Ai='ASCII';Ah='content';Ag='contentType';Af=':configuration';Ae='configuration-handling';Ad='script';Ac='hash-value';Ab='hash-algorithm';Aa='address';AZ='referenced-definition';AY='not';AX='match-criteria';AW='matched-response';AE='post-configuration-script';AD='configuration';AC='pre-configuration-script';AB='os-version';AA='os-name';A9='trust-anchor';A8='port';A7='bootstrap-server';A6='ietf-sztp-conveyed-info:redirect-information';A5='value';A4='response-manager';v='image-verification';u='download-uri';t='boot-image';s='selected-response';l=device_id;k='onboarding-information';j='key';f='reference';c='ietf-sztp-conveyed-info:onboarding-information';b='redirect-information';X=request;N='response';J='managed-response';I='response-details';E='get-bootstrapping-data-event';D='conveyed-information';C=bootstrapping_log_entry;g,AF=utils.check_http_headers(X,A.supported_media_types,accept_required=_Y)
		if type(g)is web.Response:B=g;AG=AF;C[_F]=B.status;C[_D]=AG;return B
		else:assert type(g)==str;O=g;assert O!=_K;T=utils.Encoding[O.rsplit(_G,1)[1]]
		L=_C
		if X.body_exists:
			AH=await X.text();AI=utils.Encoding[X.headers[_j].rsplit(_G,1)[1]]
			try:G=A.dm.get_schema_node(_A0);L=utils.encoded_str_to_obj(AH,AI,A.dm,G)
			except Exception as Y:B=web.Response(status=400);m=_o+str(Y);B.content_type=O;H=utils.gen_rc_errors(_L,_h,error_message=m);G=A.dm4errors.get_schema_node(_A);B.text=utils.obj_to_encoded_str(H,T,A.dm4errors,G);C[_D]=H;return B
			if not _O in L:
				B=web.Response(status=400)
				if not _O in L:m=_o+_AK
				B.content_type=O;H=utils.gen_rc_errors(_L,_h,error_message=m);G=A.dm4errors.get_schema_node(_A);B.text=utils.obj_to_encoded_str(H,T,A.dm4errors,G);C[_D]=H;return B
		C[_B]={};C[_B][E]={}
		if L is _C:C[_B][E][_A3]={'no-input-passed':[_C]}
		else:C[_B][E][_A3]=L[_O]
		if A.mode!=_e:P=_A+A.dal.app_ns+':'
		else:P=_A+A.dal.app_ns+_f+l[1]+_A
		if A.mode==_S:w=P+'device'
		else:w=P+'devices/device='+l[0]
		try:R=await A.dal.handle_get_config_request(w)
		except Exception as Y:B=web.Response(status=501);B.content_type=_m;H=utils.gen_rc_errors(_X,_l,error_message='Unhandled exception: '+str(Y));B.text=utils.enc_rc_errors('json',H);return B
		assert R!=_C;assert A.dal.app_ns+_k in R;R=R[A.dal.app_ns+_k]
		if A.mode!=_S:R=R[0]
		if A4 not in R or AW not in R[A4]:B=web.Response(status=404);B.content_type=O;H=utils.gen_rc_errors(_X,_p,error_message='No responses configured.');G=A.dm4errors.get_schema_node(_A);B.text=utils.obj_to_encoded_str(H,T,A.dm4errors,G);C[_D]=H;C[_B][E][s]='no-responses-configured';return B
		F=_C
		for h in R[A4][AW]:
			if not AX in h:F=h;break
			if L is _C:continue
			for Q in h[AX]['match']:
				if Q[j]not in L[_O]:break
				if'present'in Q:
					if AY in Q:
						if Q[j]in L[_O]:break
					elif Q[j]not in L[_O]:break
				elif A5 in Q:
					if AY in Q:
						if Q[A5]==L[_O][Q[j]]:break
					elif Q[A5]!=L[_O][Q[j]]:break
				else:raise NotImplementedError("Unrecognized 'match' expression.")
			else:F=h;break
		if F is _C or'none'in F[N]:
			if F is _C:C[_B][E][s]='no-match-found'
			else:C[_B][E][s]=F[_I]+" (explicit 'none')"
			B=web.Response(status=404);B.content_type=O;H=utils.gen_rc_errors(_X,_p,error_message='No matching responses configured.');G=A.dm4errors.get_schema_node(_A);B.text=utils.obj_to_encoded_str(H,T,A.dm4errors,G);C[_D]=H;return B
		C[_B][E][s]=F[_I];C[_B][E][I]={J:{}}
		if D in F[N]:
			C[_B][E][I][J]={D:{}};M={}
			if _E in F[N][D]:
				C[_B][E][I][J][D]={_E:{}};assert f in F[N][D][_E];n=F[N][D][_E][f];C[_B][E][I][J][D][_E][_I]=n;S=await A.dal.handle_get_config_request(P+'dynamic-callouts/dynamic-callout='+n);assert n==S[A.dal.app_ns+_H][0][_I];C[_B][E][I][J][D][_E][_q]=S[A.dal.app_ns+_H][0][_q];Z={}
				if A.mode!=_S:Z[_r]=l[0]
				else:Z[_r]=_AL
				Z[_AM]=X.remote;x=X.transport.get_extra_info(_A2)
				if x:
					y=x.getpeercert(_Y)
					if y:Z[_AN]=y
				if L:Z[_AO]=L
				if _b in S[A.dal.app_ns+_H][0]:
					C[_B][E][I][J][D][_E][_A4]=_b;z=S[A.dal.app_ns+_H][0][_b][_s];A0=S[A.dal.app_ns+_H][0][_b][_t];C[_B][E][I][J][D][_E][_AP]={_s:z,_t:A0};C[_B][E][I][J][D][_E][_c]={}
					if _u in S[A.dal.app_ns+_H][0]:A1=S[A.dal.app_ns+_H][0][_u]
					else:A1=_C
					K=_C
					try:K=A.nvh.plugins[z][_AQ][A0](Z,A1)
					except Exception as Y:C[_B][E][I][J][D][_E][_c][_AR]=str(Y);B=web.Response(status=500);B.content_type=O;H=utils.gen_rc_errors(_X,_l,error_message='Server encountered an error while trying to generate a response: '+str(Y));G=A.dm4errors.get_schema_node(_A);B.text=utils.obj_to_encoded_str(H,T,A.dm4errors,G);C[_D]=H;return B
					assert K and type(K)==dict
					if _J in K:
						assert len(K[_J][_M])==1
						if any((A==K[_J][_M][0][_N]for A in(_a,_AS,_AT,_AU,_AV,_AW,_Z,_AX,_h))):B=web.Response(status=400)
						elif any((A==K[_J][_M][0][_N]for A in _A1)):B=web.Response(status=403)
						elif any((A==K[_J][_M][0][_N]for A in(_AY,_AZ,_Aa,_Ab,_p))):B=web.Response(status=409)
						elif any((A==K[_J][_M][0][_N]for A in(_Ac,_Ad,_Ae))):B=web.Response(status=500)
						elif any((A==K[_J][_M][0][_N]for A in _l)):B=web.Response(status=501)
						else:raise NotImplementedError(_Af+K[_J][_M][0][_N])
						B.content_type=O;H=K;G=A.dm4errors.get_schema_node(_A);B.text=utils.obj_to_encoded_str(H,T,A.dm4errors,G);C[_D]=K;C[_B][E][I][J][D][_E][_c][_v]=_Ag;return B
					else:C[_B][E][I][J][D][_E][_c][_v]='Returning conveyed information provided by callback.'
				elif _A5 in S[A.dal.app_ns+_H][0]:C[_B][E][I][J][D][_E][_A4]=_w;raise NotImplementedError('webhooks callout support pending!')
				else:raise NotImplementedError('unhandled dynamic callout type: '+str(S[A.dal.app_ns+_H][0]))
				M=K
			elif b in F[N][D]:
				C[_B][E][I][J][D]={b:{}};M[A6]={};M[A6][A7]=[]
				if f in F[N][D][b]:
					d=F[N][D][b][f];C[_B][E][I][J][D][b]={AZ:d};o=await A.dal.handle_get_config_request(P+'conveyed-information-responses/redirect-information-response='+d)
					for AJ in o[A.dal.app_ns+':redirect-information-response'][0][b][A7]:
						W=await A.dal.handle_get_config_request(P+'bootstrap-servers/bootstrap-server='+AJ);W=W[A.dal.app_ns+':bootstrap-server'][0];i={};i[Aa]=W[Aa]
						if A8 in W:i[A8]=W[A8]
						if A9 in W:i[A9]=W[A9]
						M[A6][A7].append(i)
				else:raise NotImplementedError('unhandled redirect-information config type: '+str(F[N][D][b]))
			elif k in F[N][D]:
				C[_B][E][I][J][D]={};M[c]={}
				if f in F[N][D][k]:
					d=F[N][D][k][f];C[_B][E][I][J][D][k]={AZ:d};o=await A.dal.handle_get_config_request(P+'conveyed-information-responses/onboarding-information-response='+d);U=o[A.dal.app_ns+':onboarding-information-response'][0][k]
					if t in U:
						AK=U[t];AL=await A.dal.handle_get_config_request(P+'boot-images/boot-image='+AK);V=AL[A.dal.app_ns+':boot-image'][0];M[c][t]={};a=M[c][t]
						if AA in V:a[AA]=V[AA]
						if AB in V:a[AB]=V[AB]
						if u in V:
							a[u]=list()
							for AM in V[u]:a[u].append(AM)
						if v in V:
							a[v]=list()
							for A2 in V[v]:p={};p[Ab]=A2[Ab];p[Ac]=A2[Ac];a[v].append(p)
					if AC in U:AN=U[AC];AO=await A.dal.handle_get_config_request(P+'scripts/pre-configuration-script='+AN);M[c][AC]=AO[A.dal.app_ns+':pre-configuration-script'][0][Ad]
					if AD in U:AP=U[AD];A3=await A.dal.handle_get_config_request(P+'configurations/configuration='+AP);M[c][Ae]=A3[A.dal.app_ns+Af][0][Ae];M[c][AD]=A3[A.dal.app_ns+Af][0]['config']
					if AE in U:AQ=U[AE];AR=await A.dal.handle_get_config_request(P+'scripts/post-configuration-script='+AQ);M[c][AE]=AR[A.dal.app_ns+':post-configuration-script'][0][Ad]
			else:raise NotImplementedError('unhandled conveyed-information type: '+str(F[N][D]))
		else:raise NotImplementedError('unhandled response type: '+str(F[N]))
		e=rfc5652.ContentInfo()
		if O==_m:e[Ag]=A.id_ct_sztpConveyedInfoJSON;e[Ah]=encode_der(json.dumps(M,indent=2),asn1Spec=univ.OctetString())
		else:assert O==_A6;e[Ag]=A.id_ct_sztpConveyedInfoXML;G=A.dm4conveyedinfo.get_schema_node(_A);assert G;AS=utils.obj_to_encoded_str(M,T,A.dm4conveyedinfo,G,strip_wrapper=_Y);e[Ah]=encode_der(AS,asn1Spec=univ.OctetString())
		AT=encode_der(e,rfc5652.ContentInfo());q=base64.b64encode(AT).decode(Ai);AU=base64.b64decode(q);AV=base64.b64encode(AU).decode(Ai);assert q==AV;r={};r[Aj]={};r[Aj][D]=q;B=web.Response(status=200);B.content_type=O;G=A.dm.get_schema_node(_A0);B.text=utils.obj_to_encoded_str(r,T,A.dm,G);return B
	async def _handle_report_progress_rpc(A,device_id,request,bootstrapping_log_entry):
		m='remote-port';l='webhook-results';e='tcp-client-parameters';d='encoding';W='http';V=device_id;L=request;E='report-progress-event';C=bootstrapping_log_entry;S,f=utils.check_http_headers(L,A.supported_media_types,accept_required=_g)
		if type(S)is web.Response:B=S;g=f;C[_F]=B.status;C[_D]=g;return B
		else:assert type(S)==str;J=S
		if J!=_K:O=utils.Encoding[J.rsplit(_G,1)[1]]
		if not L.body_exists:
			M='RPC "input" node missing (required for "report-progress").';B=web.Response(status=400);B.content_type=J
			if B.content_type==_K:B.text=M
			else:F=utils.gen_rc_errors(_L,_a,error_message=M);G=A.dm4errors.get_schema_node(_A);B.text=utils.obj_to_encoded_str(F,O,A.dm4errors,G)
			C[_D]=B.text;return B
		h=utils.Encoding[L.headers[_j].rsplit(_G,1)[1]];i=await L.text()
		try:G=A.dm.get_schema_node(_AJ);P=utils.encoded_str_to_obj(i,h,A.dm,G)
		except Exception as K:B=web.Response(status=400);M=_o+str(K);B.content_type=J;F=utils.gen_rc_errors(_L,_h,error_message=M);G=A.dm4errors.get_schema_node(_A);B.text=utils.obj_to_encoded_str(F,O,A.dm4errors,G);C[_D]=F;return B
		if not _O in P:
			B=web.Response(status=400)
			if not _O in P:M=_o+_AK
			B.content_type=J;F=utils.gen_rc_errors(_L,_h,error_message=M);G=A.dm4errors.get_schema_node(_A);B.text=utils.obj_to_encoded_str(F,O,A.dm4errors,G);C[_D]=F;return B
		C[_B]={};C[_B][E]={};C[_B][E][_A3]=P[_O];C[_B][E][_E]={}
		if A.mode==_S or A.mode==_i:Q=_A+A.dal.app_ns+':preferences/outbound-interactions/relay-progress-report-callout'
		elif A.mode==_e:Q=_A+A.dal.app_ns+_f+V[1]+'/preferences/outbound-interactions/relay-progress-report-callout'
		try:j=await A.dal.handle_get_config_request(Q)
		except Exception as K:C[_B][E][_E]['no-callout-configured']=[_C]
		else:
			T=j[A.dal.app_ns+':relay-progress-report-callout'];C[_B][E][_E][_I]=T
			if A.mode==_S or A.mode==_i:Q=_A+A.dal.app_ns+':dynamic-callouts/dynamic-callout='+T
			elif A.mode==_e:Q=_A+A.dal.app_ns+_f+V[1]+'/dynamic-callouts/dynamic-callout='+T
			H=await A.dal.handle_get_config_request(Q);assert T==H[A.dal.app_ns+_H][0][_I];C[_B][E][_E][_q]=H[A.dal.app_ns+_H][0][_q];N={}
			if A.mode!=_S:N[_r]=V[0]
			else:N[_r]=_AL
			N[_AM]=L.remote;X=L.transport.get_extra_info(_A2)
			if X:
				Y=X.getpeercert(_Y)
				if Y:N[_AN]=Y
			if P:N[_AO]=P
			if _b in H[A.dal.app_ns+_H][0]:
				C[_B][E][_E][_A4]=_b;Z=H[A.dal.app_ns+_H][0][_b][_s];a=H[A.dal.app_ns+_H][0][_b][_t];C[_B][E][_E][_AP]={_s:Z,_t:a};C[_B][E][_E][_c]={}
				if _u in H[A.dal.app_ns+_H][0]:b=H[A.dal.app_ns+_H][0][_u]
				else:b=_C
				D=_C
				try:D=A.nvh.plugins[Z][_AQ][a](N,b)
				except Exception as K:C[_B][E][_E][_c][_AR]=str(K);B=web.Response(status=500);B.content_type=J;F=utils.gen_rc_errors(_X,_l,error_message='Server encountered an error while trying to process the progress report: '+str(K));G=A.dm4errors.get_schema_node(_A);B.text=utils.obj_to_encoded_str(F,O,A.dm4errors,G);C[_D]=F;return B
				if D:
					assert type(D)==dict;assert len(D)==1;assert _J in D;assert len(D[_J][_M])==1
					if any((A==D[_J][_M][0][_N]for A in(_a,_AS,_AT,_AU,_AV,_AW,_Z,_AX,_h))):B=web.Response(status=400)
					elif any((A==D[_J][_M][0][_N]for A in _A1)):B=web.Response(status=403)
					elif any((A==D[_J][_M][0][_N]for A in(_AY,_AZ,_Aa,_Ab,_p))):B=web.Response(status=409)
					elif any((A==D[_J][_M][0][_N]for A in(_Ac,_Ad,_Ae))):B=web.Response(status=500)
					elif any((A==D[_J][_M][0][_N]for A in _l)):B=web.Response(status=501)
					else:raise NotImplementedError(_Af+D[_J][_M][0][_N])
					B.content_type=J;F=D;G=A.dm4errors.get_schema_node(_A);B.text=utils.obj_to_encoded_str(F,O,A.dm4errors,G);C[_D]=D;C[_B][E][_E][_c][_v]=_Ag;return B
				else:C[_B][E][_E][_c][_v]='Callback returned no output (normal)'
			elif _A5 in H[A.dal.app_ns+_H][0]:
				C[_B][E][_E][l]={_w:[]}
				for I in H[A.dal.app_ns+_H][0][_A5][_w]:
					R={};R[_I]=I[_I]
					if d not in I or I[d]=='json':c=rpc_input_json
					elif I[d]=='xml':c=rpc_input_xml
					if W in I:
						U='http://'+I[W][e]['remote-address']
						if m in I[W][e]:U+=':'+str(I[W][e][m])
						U+='/relay-notification';R['uri']=U
						try:
							async with aiohttp.ClientSession()as k:B=await k.post(U,data=c)
						except aiohttp.client_exceptions.ClientConnectorError as K:R['connection-error']=str(K)
						else:
							R['http-status-code']=B.status
							if B.status==200:break
					else:assert'https'in I;raise NotImplementedError('https-based webhook is not supported yet.')
					C[_B][E][_E][l][_w].append(R)
			else:raise NotImplementedError('unrecognized callout type '+str(H[A.dal.app_ns+_H][0]))
		B=web.Response(status=204);return B